

import unittest
import argparse
import logging
import datetime
import sys
import esgfpid.utils
import globalvar

# Setup logging:
path = 'logs'
logdate = datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S')
filename = path+'/test_log_'+logdate+'.log'
esgfpid.utils.ensure_directory_exists(path)
logging.basicConfig(level=logging.DEBUG, filename=filename, filemode='w')
pikalogger = logging.getLogger('pika')
pikalogger.setLevel(logging.INFO)
#pikalogger.propagate = False


if __name__ == '__main__':
    desc = ('Test script for esgfpid library, including unit tests and integration '+
           'tests.')
    
    # Argument parsing
    parser = argparse.ArgumentParser(description=desc)
    
    # Unit or integration?
    parser.add_argument('-u','--unit', dest='unit',
                   help=('Run the unit tests?'),
                   action='store_true')
    parser.add_argument('-i','--integration', dest='inte',
                   help=('Run the integration tests?'),
                   action='store_true')
    parser.set_defaults(unit=False)
    parser.set_defaults(inte=False)

    # Modules to be tested:
    parser.add_argument('-m','--modules', metavar='mod', nargs='*',
                   help=('Which modules to test. '+
                         'Possible values: "all", "solr", "rabbit", "publish", "unpublish", '+
                         '"errata", "utils", "api", "check", "messages", "consistency", "data_cart", '+
                         '"nodemanager").'
                         'Defaults to "all".'),
                   default=['all'], action='store')

    # Synchron or asynchron?
    parser.add_argument('-noa','--no-asynchron', dest='asyn',
                   help=('Omit the asynchronous rabbit tests?'),
                   action='store_false')
    parser.add_argument('-nos','--no-synchron', dest='syn',
                   help=('Omit the synchronous rabbit tests?'),
                   action='store_false')
    parser.set_defaults(asyn=True)
    parser.set_defaults(syn=True)
    # Slow?
    parser.add_argument('-ls','--leave_out_slow', dest='leave_out',
                   help=('Leave out the slow tests?'),
                   action='store_true')
    parser.set_defaults(slow=False)
    # Open nodes?
    parser.add_argument('-open','--allow_open', dest='open_allowed',
                   help=('Can the library cope with open nodes?'),
                   action='store_true')
    parser.set_defaults(open_allowed=False)

    # Parse the args:
    param = parser.parse_args()
    print('Modules to be tested: '+ ', '.join(param.modules))

    # Inform user about skipping slow tests...
    if param.leave_out == True:
        globalvar.QUICK_ONLY = True
        print('Leaving out the slow tests (%s).' % globalvar.QUICK_ONLY)
    else:
        globalvar.QUICK_ONLY = False
    if param.open_allowed == True:
        globalvar.RABBIT_OPEN_NOT_ALLOWED = False
        print('Running tests for open RabbitMQ nodes, too (%s).' % globalvar.QUICK_ONLY)
    else:
        globalvar.RABBIT_OPEN_NOT_ALLOWED = True


    # Collect tests:
    verbosity = 5
    descriptions = 0
    print('\nCollecting tests:')
    tests_to_run = []
    numtests = 0

    if param.unit:
    
        if 'utils' in param.modules or 'all' in param.modules:

            from testcases.util_tests import UtilsTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(UtilsTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

            from testcases.util_logging_tests import UtilsLoggingTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(UtilsLoggingTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

            from testcases.errormessage_util_tests import ErrorMessageUtilsTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(ErrorMessageUtilsTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

        if 'messages' in param.modules or 'all' in param.modules:

            from testcases.messages_tests import MessageCreationTestcase
            tests = unittest.TestLoader().loadTestsFromTestCase(MessageCreationTestcase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

        if 'check' in param.modules or 'all' in param.modules:

            from testcases.check_tests import CheckTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(CheckTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

        if 'data_cart' in param.modules or 'all' in param.modules:

            from testcases.data_cart_tests import DataCartTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(DataCartTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

        if 'solr' in param.modules or 'all' in param.modules:

            from testcases.solr.solr_utils_tests import SolrUtilsTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(SolrUtilsTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

            from testcases.solr.solr_tests import SolrTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(SolrTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

            from testcases.solr.solr_task1_tests import SolrTask1TestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(SolrTask1TestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

            from testcases.solr.solr_task2_tests import SolrTask2TestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(SolrTask2TestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

            from testcases.solr.solr_server_tests import SolrServerConnectorTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(SolrServerConnectorTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

        if 'nodemanager' in param.modules or 'rabbit' in param.modules or 'all' in param.modules:

            from testcases.rabbit.nodemanager_tests import NodemanagerTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(NodemanagerTestCase)
            tests_to_run.append(tests)
            numtests += tests.countTestCases()

        if 'rabbit' in param.modules or 'all' in param.modules:

            from testcases.rabbit.rabbit_api_tests import RabbitTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(RabbitTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n
            
            from testcases.rabbit.rabbitutil_tests import RabbitUtilsTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(RabbitUtilsTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

            if param.syn:

                from testcases.rabbit.syn.rabbit_synchronous_tests import RabbitConnectorTestCase
                tests = unittest.TestLoader().loadTestsFromTestCase(RabbitConnectorTestCase)
                tests_to_run.append(tests)
                n = tests.countTestCases()
                numtests += n

            if param.asyn:

                # This tests the API and the thread
                from testcases.rabbit.asyn.rabbit_asynchronous_tests import RabbitAsynConnectorTestCase
                tests = unittest.TestLoader().loadTestsFromTestCase(RabbitAsynConnectorTestCase)
                tests_to_run.append(tests)
                numtests += tests.countTestCases()

                # Confirmer is pretty isolated and easy to test.
                from testcases.rabbit.asyn.thread_confirmer_tests import ThreadConfirmerTestCase
                tests = unittest.TestLoader().loadTestsFromTestCase(ThreadConfirmerTestCase)
                tests_to_run.append(tests)
                numtests += tests.countTestCases()

                # Feeder needs some mocking...
                from testcases.rabbit.asyn.thread_feeder_tests import ThreadFeederTestCase
                tests = unittest.TestLoader().loadTestsFromTestCase(ThreadFeederTestCase)
                tests_to_run.append(tests)
                numtests += tests.countTestCases()

                # Returner needs some mocking...
                from testcases.rabbit.asyn.thread_returner_tests import ThreadReturnerTestCase
                tests = unittest.TestLoader().loadTestsFromTestCase(ThreadReturnerTestCase)
                tests_to_run.append(tests)
                numtests += tests.countTestCases()
            
                # Shutter needs some mocking...
                from testcases.rabbit.asyn.thread_shutter_tests import ThreadShutterTestCase
                tests = unittest.TestLoader().loadTestsFromTestCase(ThreadShutterTestCase)
                tests_to_run.append(tests)
                numtests += tests.countTestCases()

                # Builder needs lots of mocking
                from testcases.rabbit.asyn.thread_builder_tests import ThreadBuilderTestCase
                tests = unittest.TestLoader().loadTestsFromTestCase(ThreadBuilderTestCase)
                tests_to_run.append(tests)
                numtests += tests.countTestCases()

        if 'errata' in param.modules or 'all' in param.modules:

            from testcases.errata_tests import ErrataTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(ErrataTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

        if 'unpublish' in param.modules or 'all' in param.modules:

            from testcases.unpublication_tests import UnpublicationTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(UnpublicationTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

        if 'publish' in param.modules or 'all' in param.modules:

            from testcases.publish_tests import PublishTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(PublishTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

        if 'api' in param.modules or 'all' in param.modules:

            from testcases.connector_tests import ConnectorTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(ConnectorTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

        if 'consistency' in param.modules or 'all' in param.modules:

            from testcases.consistency_tests import ConsistencyTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(ConsistencyTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

    if param.inte:

        if 'rabbit' in param.modules or 'all' in param.modules:

            if param.asyn:
                from testcases.integration_rabbit_tests_asynchron import RabbitIntegrationTestCase
            else:
                from testcases.integration_rabbit_tests_synchron import RabbitIntegrationTestCase

            tests = unittest.TestLoader().loadTestsFromTestCase(RabbitIntegrationTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

        if 'solr' in param.modules or 'all' in param.modules:

            from testcases.integration_solr_tests import SolrIntegrationTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(SolrIntegrationTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

            from testcases.integration_solr_server_tests import SolrServerIntegrationTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(SolrServerIntegrationTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

            from testcases.integration_solr_task1_tests import SolrTask1IntegrationTestCase
            tests = unittest.TestLoader().loadTestsFromTestCase(SolrTask1IntegrationTestCase)
            tests_to_run.append(tests)
            n = tests.countTestCases()
            numtests += n

    # Run all of them:
    print('Run '+str(numtests)+' tests.')
    test_suites = unittest.TestSuite(tests_to_run)
    print('\nStarting tests:')
    logging.info("Starting tests:")
    test_result = unittest.TextTestRunner(descriptions=descriptions, verbosity=verbosity).run(test_suites)
    sys.exit(not test_result.wasSuccessful())

    # Run with:
    # python -m coverage run main_test_script.py -u
    # python -m coverage html



import unittest
import mock
import logging
import json
import tests.utils as utils
from tests.utils import compare_json_return_errormessage as error_message

from esgfpid.assistant.errata import ErrataAssistant



# Logging
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import *
import resources.TESTVALUES as TESTHELPERS

'''
Unit tests for esgfpid.assistant.errata.

This module needs a coupler to work: It forwards calls to
RabbitMQ to the coupler.

For the tests, we use a real coupler object that has a
mocked RabbitMQ connection, and whose solr communication
is simply switched off (not mocked), as it is never used
for errata operations anyway.
'''
class ErrataTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    def test_init_ok(self):

        # Preparations
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
  
        # Run code to be tested:
        assistant = ErrataAssistant(
            prefix=PREFIX_NO_HDL,
            coupler=testcoupler
        )

        # Check result:
        self.assertIsInstance(assistant, ErrataAssistant, 'Constructor fail.')

    def test_add_errata_id_one_as_string(self):

        # Preparations
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        assistant = ErrataAssistant(
            prefix=PREFIX_NO_HDL,
            coupler=testcoupler
        )

        # Run code to be tested:
        assistant.add_errata_ids(
            drs_id=DRS_ID,
            version_number=DS_VERSION,
            errata_ids=ERRATA
        )

        # Check result:
        expected_rabbit_task = {
            "handle":DATASETHANDLE_HDL,
            "operation": "add_errata_ids",
            "message_timestamp":"anydate",
            "errata_ids":[ERRATA],
            "ROUTING_KEY": PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.errata-add',
            "drs_id":DRS_ID,
            "version_number":DS_VERSION
        }
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler)
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_add_errata_id_one_as_list(self):

        # Preparations
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        assistant = ErrataAssistant(
            prefix=PREFIX_NO_HDL,
            coupler=testcoupler
        )

        # Run code to be tested:
        assistant.add_errata_ids(
            drs_id=DRS_ID,
            version_number=DS_VERSION,
            errata_ids=[ERRATA]
        )

        # Check result:
        expected_rabbit_task = {
            "handle":DATASETHANDLE_HDL,
            "operation": "add_errata_ids",
            "message_timestamp":"anydate",
            "errata_ids":[ERRATA],
            "ROUTING_KEY": PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.errata-add',
            "drs_id":DRS_ID,
            "version_number":DS_VERSION
        }
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler)
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_add_errata_id_several(self):

        # Preparations
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        assistant = ErrataAssistant(
            prefix=PREFIX_NO_HDL,
            coupler=testcoupler
        )

        # Run code to be tested:
        assistant.add_errata_ids(
            drs_id=DRS_ID,
            version_number=DS_VERSION,
            errata_ids=ERRATA_SEVERAL
        )

        # Check result:
        expected_rabbit_task = {
            "handle":DATASETHANDLE_HDL,
            "operation": "add_errata_ids",
            "message_timestamp":"anydate",
            "errata_ids":ERRATA_SEVERAL,
            "ROUTING_KEY": PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.errata-add',
            "drs_id":DRS_ID,
            "version_number":DS_VERSION
        }
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler)
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))

    def test_add_errata_id_missing_args(self):

        # Preparations
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        assistant = ErrataAssistant(
            prefix=PREFIX_NO_HDL,
            coupler=testcoupler
        )

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.ArgumentError):
            assistant.add_errata_ids(
                drs_id=DRS_ID,
                version_number=DS_VERSION
            )

    def test_remove_errata_id_several(self):

        # Preparations
        testcoupler = TESTHELPERS.get_coupler(solr_switched_off=True)
        TESTHELPERS.patch_with_rabbit_mock(testcoupler)
        assistant = ErrataAssistant(
            prefix=PREFIX_NO_HDL,
            coupler=testcoupler
        )

        # Run code to be tested:
        assistant.remove_errata_ids(
            drs_id=DRS_ID,
            version_number=DS_VERSION,
            errata_ids=ERRATA_SEVERAL
        )

        # Check result:
        expected_rabbit_task = {
            "handle":DATASETHANDLE_HDL,
            "operation": "remove_errata_ids",
            "message_timestamp":"anydate",
            "errata_ids":ERRATA_SEVERAL,
            "ROUTING_KEY": PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.errata-rem',
            "drs_id":DRS_ID,
            "version_number":DS_VERSION
        }
        received_rabbit_task = TESTHELPERS.get_received_message_from_rabbitmock(testcoupler)
        same = utils.is_json_same(expected_rabbit_task, received_rabbit_task)
        self.assertTrue(same, error_message(expected_rabbit_task, received_rabbit_task))


import re
# Natural sorting
# From Stack Overflow: http://stackoverflow.com/questions/5967500/how-to-correctly-sort-a-string-with-a-number-inside
# We need this, as priorities are given as integers, but we need them
# as strings to use them as dictionary keys, 

def atoi(text):
    return int(text) if text.isdigit() else text

def natural_keys(text):
    '''
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    '''
    return [ atoi(c) for c in re.split('(\d+)', text) ]


class ConnectionNotReady(Exception):

    def __init__(self, custom_message=None):
        self.msg = 'Connection to rabbit not ready.'
        self.custom_message = custom_message

        if self.custom_message is not None:
            self.msg += ': '+self.custom_message
        self.msg += '.'

        super(self.__class__, self).__init__(self.msg)

class OperationNotAllowed(Exception):

    def __init__(self, custom_message=None, op=None):

        # Operation:
        if op is None:
            op=''
        else:
            op='"%s" ' % op

        # Message
        self.msg = 'Operation %snot allowed right now' % op

        # Append custom message
        if custom_message is not None:
            self.msg += ': '+custom_message
        self.msg += '.'

        super(self.__class__, self).__init__(self.msg)

class UnknownServerResponse(Exception):

    def __init__(self, custom_message=None):
        self.msg = 'Cannot understand RabbitMQ\'s response'
        self.custom_message = custom_message

        if self.custom_message is not None:
            self.msg += ': '+self.custom_message
        self.msg += '.'

        super(self.__class__, self).__init__(self.msg)


'''
This module represents the API of the ESGF PID module.

Author: Merret Buurman (DKRZ), 2015-2016

'''

import logging
import esgfpid.assistant.publish
import esgfpid.assistant.unpublish
import esgfpid.assistant.errata
import esgfpid.assistant.datacart
import esgfpid.defaults
import esgfpid.exceptions
import esgfpid.coupling
import esgfpid.utils
from esgfpid.utils import loginfo, logdebug, logwarn
import esgfpid.check

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())


class Connector(object):
    '''
    This class provides the main functionality for the ESGF PID
    module. 

    Author: Merret Buurman (DKRZ), 2015-2016
    '''

    def __init__(self, **args):
        '''
        Create a connector object with the necessary config
        to connect to a RabbitMQ messaging server and perform
        PID creations/updates.

        The arguments have to be passed as named parameters.
        Please contact your ESGF index node or CDNOT for this
        information.

        Some of the arguments are needed for making connections
        from this library to RabbitMQ or to solr. Other arguments
        are only passed on to the consuming servlet inside the
        messages.

        :param handle_prefix: Mandatory. The handle prefix (as a
            string) for the handles to be created/updates. This
            has to match the handle prefix that the message queue
            consuming servlet has write access to. In CMIP6, this
            is "21.14100".
        
        :param messaging_service_exchange_name: Mandatory. The
            name of the messaging exchange that will forward the
            messages to a specific queue.

        :param messaging_service_credentials: Mandatory. List of
            dictionaries with credentials for the RabbitMQ nodes.
            Each needs to have the entries: "user", "password", "url".
            They may have an integer "priority" too. If two nodes have
            the same priority, the library chooses randomly between
            them. They also may have a "vhost" (RabbitMQ virtual host),
            a "port" and a boolean "ssl_enabled". Please refer to pika's
            documentation
            (http://pika.readthedocs.io/en/latest/modules/parameters.html).
            Dictionaries for 'open nodes' do not need a password
            to be provided. Open nodes are only used if no more
            other nodes are available. Note: Open nodes are no longer
            supported.

        :param message_service_synchronous: Optional. Boolean to
            define if the connection to RabbitMQ and the message
            sending should work in synchronous mode. Defaults to
            the value defined in defaults.py.

        :param data_node: Mandatory/Optional.

            (Mandatory for publications and unpublications,
            ignored for any other usage of the library. No default.)

            The data node (host name) at which (un)publication takes
            place. This will be included in the handle records. Trailing
            slashes are removed.

            Used during publication and unpublication (modules
            assistant.publish and assistant.unpublish):

            * Publication: Used to construct the file data URL (together
              with thredds service path and file publish path). Sent along
              in rabbit message. Used for consistency check, if solr use 
              is enabled.
            * Unpublication: Sent along in rabbit message.
        
        :param thredds_service_path: Mandatory for publications,
            ignored for any other usage of the library. No default.
            The thredds service path where the files of a publication
            reside. Will be combined with files' publish path and data
            node to form the files' data access URLs.

        :param solr_url: Optional. The URL of the solr to be uses by this
            library for the dataset consistency check. No default. If not provided,
            the check is not done.
            Note: This is currently switched off for performance reasons.

        :param solr_https_verify: Optional flag to indicate whether
            requests to solr should verify the SSL certificate.
            Please see documentation of requests library: http://docs.python-requests.org/en/master/user/advanced/

        :param disable_insecure_request_warning: Optional flag (only for
            use during testing). If True, warnings are not printed during
            insecure SSL requests.
            Important: This is not passed through to the solr module, so
            that switching off the warnings is not possible. It can only
            be passed directly to the solr module during tests.

        :param solr_switched_off: Optional flag to tell if the solr module
            should be switched off. In that case, no connections to solr
            are made. 

        :param consumer_solr_url: Optional. URL of a solr instance that
            is to be used by the consumer (e.g. for finding versions), *not*
            by this library.

        :param test_publication: Optional flag. If True, the
            handles that are created are test handles
            that will be overwritten by real publications. Also,
            test publications cannot update real handles.

        :returns: An instance of the connector, configured for one 
            data node, and for connection with a specific RabbitMQ node.

        '''
        LOGGER.debug(40*'-')
        LOGGER.debug('Creating PID connector object ..')
        self.__check_presence_of_args(args)
        self.__check_rabbit_credentials_completeness(args)
        self.__define_defaults_for_optional_args(args)
        self.__store_some_args(args)
        self.__throw_error_if_prefix_not_in_list()
        esgfpid.utils.routingkeys.add_prefix_to_routing_keys(self.prefix)
        self.__coupler = esgfpid.coupling.Coupler(**args)
        loginfo(LOGGER, 'Created PID connector.')

    def __check_presence_of_args(self, args):
        mandatory_args = [
            'messaging_service_credentials',
            'messaging_service_exchange_name',
            'handle_prefix'
        ]
        optional_args = [
            'data_node',
            'thredds_service_path',
            'test_publication',
            'solr_url',
            'solr_https_verify',
            'disable_insecure_request_warning',
            'solr_switched_off',
            'consumer_solr_url',
            'message_service_synchronous'
        ]
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)

    def __define_defaults_for_optional_args(self, args):
        
        if 'data_node' not in args or args['data_node'] is None:
            ''' May be None, only needed for some operations.
                If it is needed, its presence is checked later. '''
            args['data_node'] = None

        if 'thredds_service_path' not in args or args['thredds_service_path'] is None:
            ''' May be None, only needed for some operations.
                If it is needed, its presence is checked later. '''
            args['thredds_service_path'] = None

        if 'test_publication' not in args or args['test_publication'] is None:
            args['test_publication'] = False

        if 'solr_url' not in args or args['solr_url'] is None:
            args['solr_url'] = None
            args['solr_switched_off'] = True

        if 'solr_switched_off' not in args or args['solr_switched_off'] is None:
            args['solr_switched_off'] = False

        if 'solr_https_verify' not in args or args['solr_https_verify'] is None:
            args['solr_https_verify'] = esgfpid.defaults.SOLR_HTTPS_VERIFY_DEFAULT

        if 'disable_insecure_request_warning' not in args or args['disable_insecure_request_warning'] is None:
            args['disable_insecure_request_warning'] = False

        if 'message_service_synchronous' not in args or args['message_service_synchronous'] is None:
            args['message_service_synchronous'] = not esgfpid.defaults.RABBIT_IS_ASYNCHRONOUS

        if 'consumer_solr_url' not in args or args['consumer_solr_url'] is None:
            args['consumer_solr_url'] = None

    def __check_rabbit_credentials_completeness(self, args):
        for credentials in args['messaging_service_credentials']:

            if not isinstance(credentials, dict):
                errmsg = 'Credentials for each RabbitMQ node should be a dictionary. Got %s.' % type(credentials)
                raise esgfpid.exceptions.ArgumentError(errmsg)

            # Mandatory:
            self.__check_presence_and_type('url', credentials, str)
            self.__check_presence_and_type('user', credentials, str)
            self.__check_presence_and_type('password', credentials, str) # If you want open nodes to be enabled again, remove this!
            
            # Optional:
            self.__check_and_adapt_type_if_exists('password', credentials, str)
            self.__check_and_adapt_type_if_exists('vhost', credentials, str)
            self.__check_and_adapt_type_if_exists('port', credentials, int)
            self.__check_and_adapt_type_if_exists('ssl_enabled', credentials, bool)

    def __check_presence_and_type(self, attname, credentials, desiredtype):
        self.__check_presence(attname, credentials)
        self.__check_and_adapt_type_if_exists(attname, credentials, desiredtype)

    def __check_presence(self, attname, credentials):
        if attname not in credentials:
            rabbitname_for_errmsg = '(not specified)'
            if 'url' in credentials:
                rabbitname_for_errmsg = credentials['url']
            errmsg = 'Missing %s for messaging service "%s"!' % (attname, rabbitname_for_errmsg)
            raise esgfpid.exceptions.ArgumentError(errmsg)

    def __check_and_adapt_type_if_exists(self, attname, credentials, desiredtype):
        if attname in credentials:

            # Empty string to None:
            if credentials[attname] == '':
                credentials[attname] = None

            # List to object:
            if type(credentials[attname]) == type([]) and len(credentials[attname]) == 1:
                credentials[attname] = credentials[attname][0]

            # Don't check if None:
            if credentials[attname] is None:
                pass

            # Check type:
            elif not isinstance(credentials[attname], desiredtype):

                # Try conversion:
                try:
                    credentials[attname] = self.__try_conversion(credentials[attname], desiredtype)

                except ValueError as e:
                    errmsg = ('Wrong type of messaging service %s (%s). Expected %s, got %s, conversion failed.' % 
                        (attname, credentials[attname], desiredtype, type(credentials[attname])))
                    raise esgfpid.exceptions.ArgumentError(errmsg)

    def __try_conversion(self, value, desiredtype):
        if desiredtype == bool:
            if isinstance(value, str):
                if str.lower(value) == 'true':
                    return True
                elif str.lower(value) == 'false':
                    return False
            raise ValueError()
        if desiredtype == str:
            #return str(value)
            raise ValueError('Not transforming booleans')
        if desiredtype == int:
            return int(value)
        else:
            return desiredtype(value)

    '''
    These are not (only) needed during initialisation, but
    (also) later on.
    '''
    def __store_some_args(self, args):
        self.prefix = args['handle_prefix']
        self.__thredds_service_path = args['thredds_service_path']
        self.__data_node = args['data_node'] # may be None, only needed for some assistants.
        self.__consumer_solr_url = args['consumer_solr_url'] # may be None

    def __throw_error_if_prefix_not_in_list(self):
        if self.prefix is None:
            raise esgfpid.exceptions.ArgumentError('Missing handle prefix!')
        if self.prefix not in esgfpid.defaults.ACCEPTED_PREFIXES:
            raise esgfpid.exceptions.ArgumentError('The prefix "%s" is not a valid prefix! Please check your config. Accepted prefixes: %s'
                % (self.prefix, ', '.join(esgfpid.defaults.ACCEPTED_PREFIXES)))

    def create_publication_assistant(self, **args):
        '''
        Create an assistant for a dataset that allows to make PID
        requests for the dataset and all of its files.

        :param drs_id: Mandatory. The dataset id of the dataset
            to be published.

        :param version_number: Mandatory. The version number of the
            dataset to be published.

        :param is_replica: Mandatory. Flag to indicate whether the
            dataset is a replica.

        .. note:: If the replica flag is set to False, the publication
            may still be considered a replica by the consuming servlet,
            namely if the dataset was already published at a different
            host. For this, please refer to the consumer documentation.

        :return: A publication assistant which provides all necessary
            methods to publish a dataset and its files.
        '''

        # Check args
        logdebug(LOGGER, 'Creating publication assistant..')
        mandatory_args = ['drs_id', 'version_number', 'is_replica']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        # Check if service path is given
        if self.__thredds_service_path is None:
            msg = 'No thredds_service_path given (but it is mandatory for publication)'
            logwarn(LOGGER, msg)
            raise esgfpid.exceptions.ArgumentError(msg)
        # Check if data node is given
        if self.__data_node is None:
            msg = 'No data_node given (but it is mandatory for publication)'
            logwarn(LOGGER, msg)
            raise esgfpid.exceptions.ArgumentError(msg)

        # Check if solr has access:
        if self.__coupler.is_solr_switched_off():
            pass # solr access not mandatory anymore

        # Create publication assistant
        assistant = esgfpid.assistant.publish.DatasetPublicationAssistant(
            drs_id=args['drs_id'],
            version_number=args['version_number'],
            thredds_service_path=self.__thredds_service_path,
            data_node=self.__data_node,
            prefix=self.prefix,
            coupler=self.__coupler,
            is_replica=args['is_replica'],
            consumer_solr_url=self.__consumer_solr_url # may be None
        )
        logdebug(LOGGER, 'Creating publication assistant.. done')
        return assistant


    '''
    Please see documentation of solr module (:func:`~check.check_pid_queue_availability`).
    '''
    def check_pid_queue_availability(self, **args):
        rabbit_checker = esgfpid.check.RabbitChecker(connector = self, prefix = self.prefix, **args)
        return rabbit_checker.check_and_inform()

    def unpublish_one_version(self, **args):
        '''
        Sends a PID update request for the unpublication of one version
        of a dataset currently published at the given data node.

        Either the handle or the pair of drs_id and version_number
        have to be provided, otherwise an exception will occur.

        The consumer will of course check the PID request message's
        timestamp with the timestamp of the last publication, so that
        republications in the meantime are not unpublished.

        The unpublication of the files is included in this method.

        :param dataset_handle: Optional. The handle of the dataset
            to be unpublished.

        :param drs_id: Optional. The dataset id of the dataset
            to be unpublished.

        :param version_number: Optional. The version number of
            the dataset to be unpublished.

        :raises: ArgumentError: If not enough arguments are passed
            to identify the dataset, or if no data node was specified
            during library init.

        '''

        # Check args
        optional_args = ['dataset_handle', 'drs_id', 'version_number']
        esgfpid.utils.add_missing_optional_args_with_value_none(args, optional_args)

        # Check if data node is given
        if self.__data_node is None:
            msg = 'No data_node given (but it is mandatory for unpublication)'
            logwarn(LOGGER, msg)
            raise esgfpid.exceptions.ArgumentError(msg)

        # Unpublish
        assistant = esgfpid.assistant.unpublish.AssistantOneVersion(
            drs_id = args['drs_id'],
            data_node = self.__data_node,
            prefix=self.prefix,
            coupler=self.__coupler,
            message_timestamp=esgfpid.utils.get_now_utc_as_formatted_string()
        )
        assistant.unpublish_one_dataset_version(
            dataset_handle = args['dataset_handle'],
            version_number = args['version_number']
        )
 
    def unpublish_all_versions(self, **args):
        '''
        Sends a PID update request for the unpublication of all versions
        of a dataset currently published at the given data node.

        If the library has solr access, it will try to find all the
        dataset versions and their handles from solr, and send individual
        messages for each version. Otherwise, one message is sent, and the
        queue consuming servlet has to identify the relevant versions,
        also making sure not to unpublish any versions that may have been
        republished in the meantime.

        :param drs_id: Dataset id of the dataset to be unpublished.
        :raises: ArgumentError: If the data node
                 was not provided at library initialization.

        '''

        # Check args
        mandatory_args = ['drs_id']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)

        # Check if data node is given
        if self.__data_node is None:
            msg = 'No data_node given (but it is mandatory for publication)'
            logwarn(LOGGER, msg)
            raise esgfpid.exceptions.ArgumentError(msg)

        # Check if solr has access:
        if self.__coupler.is_solr_switched_off():
            msg = 'Unpublication of all versions. Without solr access, we cannot identify the versions, so the consumer will have to take care of this.'
            logdebug(LOGGER, msg)
            #raise esgfpid.exceptions.ArgumentError('No solr access. Solr access is needed for publication. Please provide access to a solr index when initializing the library')

        # Unpublish
        assistant = esgfpid.assistant.unpublish.AssistantAllVersions(
            drs_id = args['drs_id'],
            data_node = self.__data_node,
            prefix=self.prefix,
            coupler=self.__coupler,
            message_timestamp=esgfpid.utils.get_now_utc_as_formatted_string(),
            consumer_solr_url = self.__consumer_solr_url # may be None
        )
        assistant.unpublish_all_dataset_versions()

    def add_errata_ids(self, **args):
        '''
        Add errata ids to a dataset handle record.

        To call this method, you do not need to provide the
        PID of the dataset. Instead, the PID string is derived
        from the dataset id and the version number.

        :param errata_ids: Mandatory. A list of errata ids (strings)
            to be added to the handle record.

        :param drs_id: Mandatory. The dataset id of the dataset
            to whose handle record the errata ids are to be
            added. (This is needed because the handle is found
            by making a hash over dataset id and version number).

        :param version_number: Mandatory. The version number of the
            dataset to whose handle record the errata ids are to be
            added. (This is needed because the handle is found by
            making a hash over dataset id and version number).
        '''

        # Check args:
        mandatory_args = ['drs_id', 'version_number', 'errata_ids']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        esgfpid.utils.check_noneness_of_mandatory_args(args, mandatory_args)

        # Perform metadata update
        assistant = esgfpid.assistant.errata.ErrataAssistant(
            coupler=self.__coupler,
            prefix=self.prefix
        )
        assistant.add_errata_ids(
            drs_id=args['drs_id'],
            version_number=args['version_number'],
            errata_ids=args['errata_ids']
        )

    def remove_errata_ids(self, **args):
        '''
        Remove errata ids from a dataset handle record.

        To call this method, you do not need to provide the
        PID of the dataset. Instead, the PID string is derived
        from the dataset id and the version number.

        :param errata_ids: Mandatory. A list of errata ids (strings) to
            be removed from the handle record.

        :param drs_id: Mandatory. The dataset id of the dataset
            from whose handle record the errata ids are to be
            removed. (This is needed because the handle is found
            by making a hash over dataset id and version number).

        :param version_number: Mandatory. The version number of the
            dataset from whose handle record the errata ids are to be
            removed. (This is needed because the handle is found by
            making a hash over dataset id and version number).

        '''

        # Check args:
        mandatory_args = ['drs_id', 'version_number', 'errata_ids']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        esgfpid.utils.check_noneness_of_mandatory_args(args, mandatory_args)

        # Perform metadata update
        assistant = esgfpid.assistant.errata.ErrataAssistant(
            coupler=self.__coupler,
            prefix=self.prefix
        )
        assistant.remove_errata_ids(
            drs_id=args['drs_id'],
            version_number=args['version_number'],
            errata_ids=args['errata_ids']
        )

    def create_data_cart_pid(self, dict_of_drs_ids_and_pids):
        '''
        Create a handle record for a data cart (a custom set of datasets).

        The handle string is made of the prefix passed to tbe library,
        and a hash over all the dataset ids in the cart. This way, if exactly
        the same set of datasets is passed several times, the same handle
        record is created, instead of making a new one.

        :param dict_of_drs_ids_and_pids: A dictionary of all dataset ids
            and their pid strings. If a dataset has no (known) PID, use
            "None".

        :return: The handle string for this data cart.
        '''
        assistant = esgfpid.assistant.datacart.DataCartAssistant(
            prefix=self.prefix,
            coupler=self.__coupler
        )
        return assistant.make_data_cart_pid(dict_of_drs_ids_and_pids)

    def start_messaging_thread(self):
        '''
        Start the parallel thread that takes care of the asynchronous
        communication with RabbitMQ.

        If PID creation/update requests are attempted before
        this was called, an exception will be raised.

        Preferably call this method as early as possible, so that
        the module has some time to build the connection before
        the first PID requests are made.
        (If PID requests are made before the connection is ready,
        they will not be lost, but pile up and sent once the connection
        is ready).

        .. important:: Please do not forget to finish the thread at the end,
            using :meth:`~esgfpid.connector.Connector.finish_messaging_thread`
            or :meth:`~esgfpid.connector.Connector.force_finish_messaging_thread`.

        '''
        self.__coupler.start_rabbit_connection()

    def finish_messaging_thread(self):
        '''
        Finish and join the parallel thread that takes care of
        the asynchronous communication with RabbitMQ.

        If some messages are still in the stack to be sent,
        or if some messages were not confirmed yet, this method
        blocks and waits for some time while it iteratively
        checks for message confirmation.

        Currently, it waits up to 5 seconds: It checks up to
        11 times, waiting 0.5 seconds in between - these
        values can be configured in the defaults module).
        '''
        self.__coupler.finish_rabbit_connection()

    def force_finish_messaging_thread(self):
        '''
        Finish and join the parallel thread that takes care of
        the asynchronous communication with RabbitMQ.

        This method does not wait for any pending messages.
        Messages that are not sent yet are lost. Messages that
        are not confirmed yet are probably not lost, but their
        receival is not guaranteed.

        Note:
        The rabbit module keeps a copy of all unsent and
        unconfirmed messages, so they could be resent in
        a later connection. It would also be easy to expose
        a method for the library caller to retrieve those
        messages, e.g. to write them into some file.


        '''
        self.__coupler.force_finish_rabbit_connection()

    def make_handle_from_drsid_and_versionnumber(self, **args):
        '''
        Create a handle string for a specific dataset, based
        on its dataset id and version number, and the prefix
        passed to the library at initializing.

        :param drs_id: The dataset id of the dataset.
        :param version_number: The version number of the dataset
            (as a string or integer, this does not matter)
        :return: A handle string (e.g. "hdl:21.14100/abcxyzfoo")
        '''
        args['prefix'] = self.prefix
        return esgfpid.utils.make_handle_from_drsid_and_versionnumber(**args)


from setuptools import setup, find_packages

# Dependencies for using the library:
dependencies = [
    'pika>=1.0.1,<=1.1.0',
    'requests>=2.22.0'
]

# Dependencies for running the tests:
test_dependencies = [
    'mock'
]

# Packages to be included for using the library:
packages = [
    'esgfpid',
    'esgfpid/assistant',
    'esgfpid/utils',
    'esgfpid/rabbit',
    'esgfpid/rabbit/synchronous',
    'esgfpid/rabbit/asynchronous',
    'esgfpid/solr',
    'esgfpid/solr/tasks'
]

# Packages to be included for running the tests:
test_packages = [
    'tests',
    'tests/resources',
    'tests/testcases',
    'tests/testcases/solr',
    'tests/testcases/rabbit',
    'tests/testcases/rabbit/syn',
    'tests/testcases/rabbit/asyn',
    'tests/utils'
]

# Load description from README.md   
# Note: The package maintainer needs pypandoc and pygments to properly convert
# the Markdown-formatted README into RestructuredText before uploading to PyPi
# See https://bitbucket.org/pypa/pypi/issues/148/support-markdown-for-readmes
try:
    import pypandoc
    long_description=pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    long_description=open('README.md').read()


setup(
    name='esgfpid',
    version='0.8.1',
    author='Merret Buurman, German Climate Computing Centre (DKRZ)',
    author_email='buurman@dkrz.de',
    url='https://github.com/IS-ENES-Data/esgf-pid',
    download_url='https://github.com/IS-ENES-Data/esgf-pid/archive/0.8.0.tar.gz',
    description='Library for sending PID requests to a rabbit messaging queue during ESGF publication.',
    long_description=long_description,
    packages=packages + test_packages,
    install_requires=dependencies,
    tests_require=test_dependencies,
    include_package_data=True,
    classifiers=[
       'Development Status :: 4 - Beta',
       'Programming Language :: Python :: 2',
       'Programming Language :: Python :: 2.7',
       'Programming Language :: Python :: 3.7',
       'License :: OSI Approved :: Apache Software License',
       'Intended Audience :: Developers',
       'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    zip_safe=False
)



import unittest
import mock
import logging
import json
import esgfpid.solr.solr
import esgfpid.solr.tasks.filehandles_same_dataset as task
import tests.resources.responsemock
import tests.utils
import tests.resources

# Logging:
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import *
import resources.TESTVALUES as TESTHELPERS

# Load some data that is needed for testing
PATH_RES = tests.utils.get_super_neighbour_directory(__file__, 'resources')
SOLR_RESPONSE = json.load(open(PATH_RES+'/solr_response.json'))

QUERY1 = {'format': 'application/solr+json', 'facets': 'tracking_id', 'limit': 0, 'distrib': False, 'dataset_id':'abc.v2016|foo.de', 'type': 'File'}
QUERY2 = {'format': 'application/solr+json', 'facets': 'tracking_id', 'limit': 0, 'distrib': False, 'query': 'dataset_id:abc.v2016|*', 'type': 'File'}


class SolrTask1TestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    def make_testtask(self):
        testsolr = TESTHELPERS.get_testsolr()
        testtask = task.FindFilesOfSameDatasetVersion(testsolr)
        return testtask

    def get_args_dict(self):
        return dict(
            drs_id = 'abc',
            version_number = '2016',
            data_node = 'foo.de',
            prefix = '123')

    def fake_solr_response(self, ids):
        resp = {
            "facet_counts": {
                "facet_fields": {
                    "tracking_id": ids
                }
            }
        }
        return resp

    # Actual tests:

    def test_init_ok(self):

        # Preparations
        testsolr = TESTHELPERS.get_testsolr()

        # Run code to be tested:
        testtask = task.FindFilesOfSameDatasetVersion(testsolr)

        # Check result
        self.assertIsInstance(testtask, task.FindFilesOfSameDatasetVersion, 'Constructor fail.')

    @mock.patch('esgfpid.solr.serverconnector.SolrServerConnector.send_query')
    def test_retrieve_file_handles_of_same_dataset_A_ok_patched(self, getpatch):
        '''
        In this test, only strategy 1 is used.
        serverconnector.send_query returns three handles on the first call.
        '''

        # Define the replacement for the patched method:
        handles = ["123/456",3,"123/234",1,"987/567",2]
        getpatch.return_value = self.fake_solr_response(handles)

        # Preparations
        task = self.make_testtask()

        # Test variables:
        args = self.get_args_dict()

        # received_handles code to be tested:
        received_handles = task.retrieve_file_handles_of_same_dataset(**args)

        # Check result:
        # Was the correct query sent?
        #expected_query = {'format': 'application/solr+json', 'facets': 'handle,tracking_id', 'limit': 0, 'distrib': False, 'dataset_id': 'abc.v2016|foo.de', 'type': 'File'}
        expected_query = QUERY1
        getpatch.assert_called_once_with(expected_query)
        # Was the response treated correctly?
        expected_handles = ['hdl:123/987/567', 'hdl:123/234', 'hdl:123/456']
        self.assertEqual(sorted(expected_handles), sorted(received_handles), 'Expected %s, but got %s' % (expected_handles, received_handles))


    @mock.patch('esgfpid.solr.serverconnector.SolrServerConnector.send_query')
    def test_retrieve_file_handles_of_same_dataset_AB_nohandles_patched(self, getpatch):
        '''
        In this test, both strategies are used.
        serverconnector.send_query returns [] on the first call,
        so the second call is issued, but this also returns [].
        '''

        # Define the replacement for the patched method:
        getpatch.return_value = self.fake_solr_response([])

        # Preparations
        task = self.make_testtask()

        # Test variables:
        args = self.get_args_dict()

        # Run code to be tested:
        received_handles = task.retrieve_file_handles_of_same_dataset(**args)

        # Check result:
        # Was the correct query sent?
        expected_query_1 = QUERY1
        expected_query_2 = QUERY2
        getpatch.assert_any_call(expected_query_1)
        getpatch.assert_called_with(expected_query_2)
        # Was the response treated correctly?
        self.assertEqual(received_handles, [], 'Expected empty list, but got: '+str(received_handles))

    @mock.patch('esgfpid.solr.serverconnector.SolrServerConnector.send_query')
    def test_retrieve_file_handles_of_same_dataset_A_nohandle_B_ok_patched(self, getpatch):
        '''
        In this test, both strategies are used.
        serverconnector.send_query returns [] on the first call,
        so the second call is issued, but this also returns [].
        '''

        # Test variables:
        args = self.get_args_dict()
        handles = ["123/456",3,"123/234",1,"987/567",2]
        
        # Define the replacement for the patched method:
        def different_mock_response_depending_on_query(query):
            if query == QUERY1:
                return self.fake_solr_response([])
            elif query == QUERY2:
                return self.fake_solr_response(handles)
            else:
                raise ValueError('Something went wrong with the test. Wrong query: '+str(query))        
        getpatch.side_effect = different_mock_response_depending_on_query

        # Preparations
        task = self.make_testtask()

        # Run code to be tested:
        received_handles = task.retrieve_file_handles_of_same_dataset(**args)

        # Check result:
        # Was the correct query sent?
        expected_query_1 = QUERY1
        expected_query_2 = QUERY2
        getpatch.assert_any_call(expected_query_1)
        getpatch.assert_called_with(expected_query_2)
        # Was the response treated correctly?
        expected_handles = ['hdl:123/987/567', 'hdl:123/234', 'hdl:123/456']
        self.assertEqual(sorted(expected_handles), sorted(received_handles), 'Expected %s, but got %s' % (expected_handles, received_handles))

    @mock.patch('esgfpid.solr.serverconnector.SolrServerConnector.send_query')
    def test_retrieve_file_handles_of_same_dataset_A_error_B_ok_patched(self, getpatch):
        '''
        In this test, both strategies are used.
        serverconnector.send_query returns [] on the first call,
        so the second call is issued, but this also returns [].
        '''

        # Test variables:
        args = self.get_args_dict()
        handles = ["123/456",3,"123/234",1,"987/567",2]
        
        # Define the replacement for the patched method:
        def different_mock_response_depending_on_query(query):
            if query == QUERY1:
                raise esgfpid.exceptions.SolrError('Whatever...')
            elif query == QUERY2:
                return self.fake_solr_response(handles)
            else:
                raise ValueError('Something went wrong with the test. Wrong query: '+str(query))        
        getpatch.side_effect = different_mock_response_depending_on_query

        # Preparations
        task = self.make_testtask()

        # Run code to be tested:
        received_handles = task.retrieve_file_handles_of_same_dataset(**args)

        # Check result:
        # Was the correct query sent?
        expected_query_1 = QUERY1
        expected_query_2 = QUERY2
        getpatch.assert_any_call(expected_query_1)
        getpatch.assert_called_with(expected_query_2)
        # Was the response treated correctly?
        expected_handles = ['hdl:123/987/567', 'hdl:123/234', 'hdl:123/456']
        self.assertEqual(sorted(expected_handles), sorted(received_handles), 'Expected %s, but got %s' % (expected_handles, received_handles))

    @mock.patch('esgfpid.solr.serverconnector.SolrServerConnector.send_query')
    def test_retrieve_file_handles_of_same_dataset_A_error_B_nohandles_patched(self, getpatch):
        '''
        In this test, both strategies are used.
        serverconnector.send_query returns [] on the first call,
        so the second call is issued, but this also returns [].
        '''

        # Test variables:
        args = self.get_args_dict()
        
        # Define the replacement for the patched method:
        def different_mock_response_depending_on_query(query):
            if query == QUERY1:
                raise esgfpid.exceptions.SolrError('Whatever...')
            elif query == QUERY2:
                return self.fake_solr_response([])
            else:
                raise ValueError('Something went wrong with the test. Wrong query: '+str(query))        
        getpatch.side_effect = different_mock_response_depending_on_query

        # Preparations
        task = self.make_testtask()

        # Run code to be tested:
        received_handles = task.retrieve_file_handles_of_same_dataset(**args)

        # Check result:
        # Was the correct query sent?
        expected_query_1 = QUERY1
        expected_query_2 = QUERY2
        getpatch.assert_any_call(expected_query_1)
        getpatch.assert_called_with(expected_query_2)
        # Was the response treated correctly?
        self.assertEqual([], received_handles, 'Expected empty list, but got %s' % received_handles)

    @mock.patch('esgfpid.solr.serverconnector.SolrServerConnector.send_query')
    def test_retrieve_file_handles_of_same_dataset_A_error_B_error_patched(self, getpatch):
        '''
        In this test, both strategies are used.
        serverconnector.send_query returns [] on the first call,
        so the second call is issued, but this also returns [].
        '''

        # Test variables:
        args = self.get_args_dict()
        
        # Define the replacement for the patched method:
        def different_mock_response_depending_on_query(query):
            if query == QUERY1:
                raise esgfpid.exceptions.SolrError('Whatever 1...')
            elif query == QUERY2:
                raise esgfpid.exceptions.SolrError('Whatever 2...')
            else:
                raise ValueError('Something went wrong with the test. Wrong query: '+str(query))        
        getpatch.side_effect = different_mock_response_depending_on_query

        # Preparations
        task = self.make_testtask()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrError) as raised:
            received_handles = task.retrieve_file_handles_of_same_dataset(**args)
        self.assertIn('Failure in both queries', repr(raised.exception))
        self.assertIn('Whatever 1', repr(raised.exception))
        self.assertIn('Whatever 2', repr(raised.exception))


    @mock.patch('esgfpid.solr.serverconnector.SolrServerConnector.send_query')
    def test_retrieve_file_handles_of_same_dataset_A_nohandle_B_error_patched(self, getpatch):
        '''
        In this test, both strategies are used.
        serverconnector.send_query returns [] on the first call,
        so the second call is issued, but this also returns [].
        '''

        # Test variables:
        args = self.get_args_dict()
        
        # Define the replacement for the patched method:
        def different_mock_response_depending_on_query(query):
            if query == QUERY1:
                return self.fake_solr_response([])
            elif query == QUERY2:
                raise esgfpid.exceptions.SolrError('Whatever 2...')
            else:
                raise ValueError('Something went wrong with the test. Wrong query: '+str(query))        
        getpatch.side_effect = different_mock_response_depending_on_query

        # Preparations
        task = self.make_testtask()

        # Run code to be tested and check exception:
        with self.assertRaises(esgfpid.exceptions.SolrError) as raised:
            received_handles = task.retrieve_file_handles_of_same_dataset(**args)
        self.assertIn('Failure in both queries', repr(raised.exception))
        self.assertIn('First query returned an empty list', repr(raised.exception))
        self.assertIn('Whatever 2', repr(raised.exception))





'''
When the library was started in the synchronous mode and
a message could not be delivered, this exception is raised.

We use an exception instead of a boolean return value, as 
the asynchronous mode cannot return any success booleans,
so we prefer to not even pretend we can return anything.
'''
class MessageNotDeliveredException(Exception):

    def __init__(self, custom_message, undelivered_message):
        self.rabbit_msg = undelivered_message
        self.msg = 'Message could not be delivered'
        self.custom_message = custom_message

        if self.custom_message is not None:
            self.msg += ': '+self.custom_message
        self.msg += '.'

        super(self.__class__, self).__init__(custom_message)

class OperationUnsupportedException(Exception):

    def __init__(self, custom_message=None):
        self.msg = 'Operation not supported at this point'
        self.custom_message = custom_message

        if self.custom_message is not None:
            self.msg += ': '+self.custom_message
        self.msg += '.'

        super(self.__class__, self).__init__(self.msg)

class InconsistentFilesetException(Exception):

    def __init__(self, custom_message=None):
        self.msg = 'Fileset is inconsistent with previously published copy'
        self.custom_message = custom_message

        if self.custom_message is not None:
            self.msg += ': '+self.custom_message
        self.msg += '.'

        super(self.__class__, self).__init__(self.msg)

class ArgumentError(Exception):

    def __init__(self, custom_message=None):
        self.msg = 'The arguments that were passed are not ok'
        self.custom_message = custom_message

        if self.custom_message is not None:
            self.msg += ': '+self.custom_message
        self.msg += '.'

        super(self.__class__, self).__init__(self.msg)

class ESGFException(Exception):

    def __init__(self, custom_message=None):
        self.msg = 'ESGF rule violation'
        self.custom_message = custom_message

        if self.custom_message is not None:
            self.msg += ': '+self.custom_message
        self.msg += '.'

        super(self.__class__, self).__init__(self.msg)

class SolrSwitchedOff(Exception):

    def __init__(self, custom_message=None):
        self.msg = 'Solr module is switched off'
        self.custom_message = custom_message

        if self.custom_message is not None:
            self.msg += ': '+self.custom_message
        self.msg += '.'

        super(self.__class__, self).__init__(self.msg)


class SolrError(Exception):

    def __init__(self, custom_message=None):
        self.msg = 'Error during communication with solr'
        self.custom_message = custom_message

        if self.custom_message is not None:
            self.msg += ': '+self.custom_message
        self.msg += '.'

        super(self.__class__, self).__init__(self.msg)

class SolrResponseError(Exception):

    def __init__(self, custom_message=None):
        self.msg = 'Error parsing solr response'
        self.custom_message = custom_message

        if self.custom_message is not None:
            self.msg += ': '+self.custom_message
        self.msg += '.'

        super(self.__class__, self).__init__(self.msg)




class StateMachine(object):
    
    def __init__(self):

        # The four main states of the state machine:
        self.__NOT_STARTED_YET = 0
        self.__WAITING_TO_BE_AVAILABLE = 1
        self.__IS_AVAILABLE = 2
        self.__IS_AVAILABLE_BUT_WANTS_TO_STOP = 3
        self.__PERMANENTLY_UNAVAILABLE = 4 # roughly corresponds to _stopping and _closing in pika usage example.
        self.__FORCE_FINISHED = 5

        # Init state machine:
        self.__state = self.__NOT_STARTED_YET

        # More detail
        self.__detail_closed_by_publisher = False # this needs a setter, as it depends on the others!
        self.detail_asked_to_force_close_by_publisher = False
        self.detail_asked_to_gently_close_by_publisher = False
        self.detail_could_not_connect = False

    #
    # Setters
    #

    ''' Called by the rabbit thread.'''
    def set_to_available(self):
        if self.is_PERMANENTLY_UNAVAILABLE() or self.is_FORCE_FINISHED():
            pass
        else:
            self.__state = self.__IS_AVAILABLE

    ''' Called by the main thread.'''
    def set_to_wanting_to_stop(self):
        if self.is_PERMANENTLY_UNAVAILABLE() or self.is_FORCE_FINISHED():
            pass
        else:
            self.__state = self.__IS_AVAILABLE_BUT_WANTS_TO_STOP

    ''' Called by the main thread.'''
    def set_to_waiting_to_be_available(self):
        if self.is_PERMANENTLY_UNAVAILABLE() or self.is_FORCE_FINISHED():
            pass
        else:
            self.__state = self.__WAITING_TO_BE_AVAILABLE

    ''' Called by the rabbit thread.'''
    def set_to_permanently_unavailable(self):
        if self.is_FORCE_FINISHED():
            pass
        else:
            self.__state = self.__PERMANENTLY_UNAVAILABLE

    def set_to_force_finished(self):
        self.__state = self.__FORCE_FINISHED

    #
    # Getters for states
    #

    def is_NOT_STARTED_YET(self):
        if self.__state == self.__NOT_STARTED_YET:
            return True
        return False

    def is_WAITING_TO_BE_AVAILABLE(self):
        if self.__state == self.__WAITING_TO_BE_AVAILABLE:
            return True
        return False

    def is_AVAILABLE(self):
        if self.__state == self.__IS_AVAILABLE:
            return True
        return False

    def is_AVAILABLE_BUT_WANTS_TO_STOP(self):
        if self.__state == self.__IS_AVAILABLE_BUT_WANTS_TO_STOP:
            return True
        return False

    def is_PERMANENTLY_UNAVAILABLE(self):
        if self.__state == self.__PERMANENTLY_UNAVAILABLE or self.__state == self.__FORCE_FINISHED:
            # Including FORCE_FINISHED is kind of a dirty hack here,
            # but otherwise I might break to many things...
            # This hack should not be necessary anymore.
            return True
        return False

    def is_FORCE_FINISHED(self):
        if self.__state == self.__FORCE_FINISHED:
            return True
        return False


    '''
    Needed by asynchronous.py to inform if messages
    are not accepted anymore.
    '''
    def get_reason_shutdown(self):
        if self.detail_could_not_connect:
            return 'Could not connect'
        elif self.__detail_closed_by_publisher or self.detail_asked_to_gently_close_by_publisher or self.detail_asked_to_force_close_by_publisher:
            return 'Was closed by publisher'
            # The former two are more important, so only if no
            # authentication exception or could-not-connect occurred,
            # we give user close as the reason.
        else:
            return 'Unknown'

    #
    # Detailed setters
    #

    def set_detail_closed_by_publisher(self):
        if self.detail_could_not_connect:
            # If the connection was already closed because of these,
            # it was not really closed by publisher!!
            pass
        else:
            self.__detail_closed_by_publisher = True

    def get_detail_closed_by_publisher(self):
        return self.__detail_closed_by_publisher


import esgfpid.exceptions

def check_presence_of_mandatory_args(args, mandatory_args):
    missing_args = []
    for name in mandatory_args:
        if name not in list(args.keys()):
            missing_args.append(name)
    if len(missing_args)>0:
        raise esgfpid.exceptions.ArgumentError('Missing mandatory arguments: '+', '.join(missing_args))
    else:
        return True


def check_noneness_of_mandatory_args(args, mandatory_args):
    empty_args = []
    for name in mandatory_args:
        if args[name] is None:
            empty_args.append(name)
    if len(empty_args)>0:
        raise esgfpid.exceptions.ArgumentError('Problem: These arguments are None: '+', '.join(empty_args))
    else:
        return True

def add_missing_optional_args_with_value_none(args, optional_args):
    for name in optional_args:
        if not name in list(args.keys()):
            args[name] = None
    return args

def find_additional_args(args, known_args_list):
    additional_args = {}
    for argname in args:
        if argname not in known_args_list:
            additional_args[argname] = args[argname]
    return additional_args




import pika
import logging
import socket
from esgfpid.utils import check_presence_of_mandatory_args
from esgfpid.utils import add_missing_optional_args_with_value_none
import esgfpid.defaults
import esgfpid.rabbit.rabbitutils
import esgfpid.utils as utils
import esgfpid.exceptions

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())


'''
Check whether the RabbitMQ instances of a connector
object are available and reachable.


Example result:
*************************************************************************************
*** PROBLEM IN SETTING UP                                                         ***
*** RABBIT MESSAGING QUEUE (PID MODULE)                                           ***
*** CONNECTION TO THE PID MESSAGING QUEUE FAILED DEFINITIVELY:                    ***
***  - host "myrabbit.de": Connection failure (wrong host or port?).              ***
***  - host "myrabbit.de": Virtual host "foobarrr" does not exist.                ***
***  - host "myrabbit.de": Authentication failure (user foobar, password foobar). ***
***  - host "myrabbit.de": Connection failure (wrong host or port?).              ***
*** PLEASE NOTIFY handle@dkrz.de AND INCLUDE THIS ERROR MESSAGE.                  ***
*************************************************************************************


:param connector: Connector object, readily configured with RabbitMQ
    instance(s).
:param print_to_console: Optional. Boolean. If True, human-readable messages
    about the failed actions will be printed to screen (actual print method.)
:param print_success_to_console: Optional. Boolean. If True, human-readable
    messages about the successful actions will be printed to screen (actual
    print method.)
:param send_message: Optional. Boolean. If True, a test message will be sent to
    a RabbitMQ instance after successful connection.
:return: String message, or None. If the check passed, returns None. Otherwise,
    returns a detailed, printable string problem message - see example above.
'''
def check_pid_queue_availability(**args):
    rabbit_checker = RabbitChecker(**args)
    return rabbit_checker.check_and_inform()

class RabbitChecker(object):

    #
    # Init
    #

    def __init__(self, **args):
        mandatory_args = ['connector']
        optional_args = ['send_message', 'print_to_console', 'print_success_to_console']
        check_presence_of_mandatory_args(args, mandatory_args)
        add_missing_optional_args_with_value_none(args, optional_args)
        self.__define_all_attributes()
        self.__fill_all_attributes(args)
        self.connection = None
        self.channel = None
        self.channel_ok = False

    def __define_all_attributes(self):
        self.__print_errors_to_console = False
        self.__print_success_to_console = False
        self.__default_log_level = logging.DEBUG
        self.__error_messages = []
        self.__nodemanager = None
        self.__current_rabbit_host = None
        self.__exchange_name = None
        self.__send_message = False
        self.__prefix = None

    def __fill_all_attributes(self, args):
        self.__nodemanager = args['connector']._Connector__coupler._Coupler__rabbit_message_sender._RabbitMessageSender__node_manager
        if args['print_to_console'] is not None and args['print_to_console'] == True:
            self.__print_errors_to_console = True
        if args['print_success_to_console'] is not None and args['print_success_to_console'] == True:
            self.__print_success_to_console = True
        if args['send_message'] is not None and args['send_message'] == True:
            self.__send_message = True
        self.__prefix = args['connector'].prefix



    #
    # Perform the checks
    #

    '''
    Check whether the RabbitMQ instances that were passed to the connector
    are still available and reachable.
    All the necessary args have to be passed on init of the RabbitChecker
    object.
    '''
    def check_and_inform(self):
        self.__loginfo('Checking config for PID module (rabbit messaging queue) ...')
        success = self.__iterate_over_all_hosts()
        msg = None
        if success:
            self.__loginfo('Config for PID module (rabbit messaging queue).. ok.')
            self.__loginfo('Successful connection to PID messaging queue at "%s".' % self.__current_rabbit_host)

        else:
            self.__loginfo('Config for PID module (rabbit messaging queue) .. FAILED!')
            msg = self.__assemble_error_message()

        if self.connection is not None:
            self.connection.close()
        return msg

    def __iterate_over_all_hosts(self):
        success = False
        print_conn = True
        print_chan = True
        self.channel_ok = False

        while True:
            try:
                if print_conn:
                    self.__loginfo(' .. checking authentication and connection ...')
                    print_conn = False
                    print_chan = True
                
                self.connection = self.__check_making_rabbit_connection()

                if print_chan:
                    self.__loginfo(' .. checking authentication and connection ... ok.')
                    self.__loginfo(' .. checking channel ...')
                    print_chan = False
                    print_conn = True

                self.channel_ok = False
                self.channel = self.__check_opening_channel(self.connection)
                self.channel_ok = True
                self.__check_exchange_existence(self.channel)

                if self.__send_message:
                    if self.__prefix is None:
                        # Cannot happen, as we get it from the connector object.
                        raise esgfpid.exceptions.ArgumentError('Missing handle prefix!')

                    self.__check_send_print_message(self.channel)

                success = True

                break # success, leave loop

            except ValueError as e:

                self.__loginfo(' .. giving this node a lower priority..')
                self.__nodemanager.set_priority_low_for_current()

                if self.__nodemanager.has_more_urls(): # stay in loop, try next host
                    utils.logtrace(LOGGER, 'Left URLs: %s', self.__nodemanager.get_num_left_urls())
                    self.__nodemanager.set_next_host()
                    self.__current_rabbit_host = self.__nodemanager.get_connection_parameters().host
                    utils.logtrace(LOGGER, 'Now trying: %s', self.__current_rabbit_host)

                else: # definitive fail, leave loop
                    break

        return success
            
    #
    # Building connections:
    #

    def __check_exchange_existence(self, channel):

        self.__exchange_name = self.__nodemanager.get_exchange_name()
        if self.__exchange_name is not None:
            try:
                self.__loginfo(' .. checking exchange ...')
                channel.exchange_declare(self.__exchange_name, passive=True)
                self.__loginfo(' .. checking exchange ... ok.')
            except (pika.exceptions.ChannelClosed) as e:
                self.__loginfo(' .. checking exchange ... failed.')
                self.__add_error_message_no_exchange()
                raise ValueError('The exchange %s does not exist on messaging service host %s' %
                    (self.__exchange_name, self.__current_rabbit_host))
        else:
            pass # No exchange name was given

    def __check_send_print_message(self, channel):
        props = pika.BasicProperties(
            delivery_mode = 2
        )
        
        # This also does the trick:
        #esgfpid.utils.routingkeys.add_prefix_to_routing_keys(self.__prefix)
        #rkey = utils.routingkeys.ROUTING_KEYS['pre_flight']
        rkey = utils.routingkeys.ROUTING_KEYS_TEMPLATES['pre_flight']
        sanitized_prefix = utils.routingkeys._sanitize_prefix(self.__prefix)
        rkey = rkey.replace('PREFIX', sanitized_prefix)
        if 'PREFIX' in rkey:
            raise ValueError('Prefix placeholder in routing key was not replaced!')

        body = 'PLEASE PRINT: Testing pre-flight check...'
        self.__loginfo(' .. checking message ...')
        try:
            res = channel.basic_publish(
                exchange=self.__nodemanager.get_exchange_name(),
                routing_key=rkey,
                body=body,
                properties=props,
                mandatory=True
            )
            self.__loginfo(' .. checking message ... ok.')
        except pika.exceptions.UnroutableError:
            self.__loginfo(' .. checking message ... failed.')
            self.__add_error_message_message_fail(rkey)
            raise ValueError('Could not send message to messaging service host %s' %
                    (self.__current_rabbit_host))


    def __check_opening_channel(self, connection):
        channel = None
        try:
            channel = self.__open_channel(connection)
            self.__loginfo(' .. checking channel ... ok.')

        except pika.exceptions.ChannelClosed:
            self.__loginfo(' .. checking channel ... FAILED.')
            self.__add_error_message_channel_closed()
            raise ValueError('Channel failed, please try next.')

        return channel

    def __open_channel(self, connection):
        channel = connection.channel()
        channel.confirm_delivery()
        return channel

    def __check_making_rabbit_connection(self):
        connection = None
        try:
            connection = self.__open_rabbit_connection()

        except pika.exceptions.ProbableAuthenticationError:
            self.__loginfo(' .. checking authentication (%s)... FAILED.' % self.__current_rabbit_host)
            self.__add_error_message_authentication_error()
            raise ValueError('Connection failed, please try next.')

        except pika.exceptions.ProbableAccessDeniedError as e:
            self.__loginfo(' .. checking access (%s)... FAILED.' % self.__current_rabbit_host)
            vh = self.__nodemanager.get_connection_parameters().virtual_host
            if ('vhost %s not found' % vh) in str(e):
                self.__add_error_message_access_denied()
            else:
                self.__add_error_message_access_denied_unclear()
            raise ValueError('Access failed, please try next.')

        except (pika.exceptions.ConnectionClosed, socket.gaierror):
            self.__loginfo(' .. checking connection (%s)... FAILED.' % self.__current_rabbit_host)
            self.__add_error_message_connection_closed()
            raise ValueError('Connection failed, please try next.')
        
        if connection is None or not connection.is_open:
            self.__loginfo(' .. checking connection (%s)... FAILED.' % self.__current_rabbit_host)
            self.__add_error_message_connection_problem()
            raise ValueError('Connection failed, please try next.')

        self.__loginfo(' .. checking authentication and connection (%s)... ok.' % self.__current_rabbit_host)
        return connection

    def __open_rabbit_connection(self):
        params = self.__nodemanager.get_connection_parameters()
        self.__current_rabbit_host = params.host
        connection = self.__pika_blocking_connection(params)
        return connection

    def __pika_blocking_connection(self, params): # this is easy to mock
        return pika.BlockingConnection(params)

    #
    # Error messages
    #

    def __add_error_message_general(self):
        self.__error_messages.insert(0,'PROBLEM IN SETTING UP')
        self.__error_messages.insert(1,'RABBIT MESSAGING QUEUE (PID MODULE)')
        self.__error_messages.insert(2, 'CONNECTION TO THE PID MESSAGING QUEUE FAILED DEFINITIVELY:')
        self.__error_messages.append('PLEASE NOTIFY handle@dkrz.de AND INCLUDE THIS ERROR MESSAGE.')

    def __add_error_message_message_fail(self, rkey):
        msg = ' - host "%s": Message failure with routing key "%s".' % (self.__current_rabbit_host, rkey)
        self.__error_messages.append(msg)

    def __add_error_message_channel_closed(self):
        msg = ' - host "%s": Channel failure.' % self.__current_rabbit_host
        self.__error_messages.append(msg)

    def __add_error_message_access_denied(self):
        vh = self.__nodemanager.get_connection_parameters().virtual_host
        msg = ' - host "%s": Virtual host "%s" does not exist.' % (self.__current_rabbit_host, vh)
        self.__error_messages.append(msg)

    def __add_error_message_access_denied_unclear(self):
        msg = ' - host "%s": Access denied.' % (self.__current_rabbit_host)
        self.__error_messages.append(msg)

    def __add_error_message_no_exchange(self):
        msg = ' - host "%s": Exchange %s does not exist.' % (self.__current_rabbit_host, self.__exchange_name)
        self.__error_messages.append(msg)

    def __add_error_message_authentication_error(self):
        msg = (' - host "%s": Authentication failure (user %s, password %s).' % (
            self.__current_rabbit_host,
            self.__nodemanager.get_connection_parameters().credentials.username,
            self.__nodemanager.get_connection_parameters().credentials.password
        ))
        self.__error_messages.append(msg)

    def __add_error_message_connection_closed(self):
        msg = ' - host "%s": Connection failure (wrong host or port?).' % self.__current_rabbit_host
        self.__error_messages.append(msg)

    def __add_error_message_connection_problem(self):
        msg = ' - host "%s": Unknown connection failure.' % self.__current_rabbit_host
        self.__error_messages.append(msg)

    #
    # Inform at the end
    #

    def __assemble_error_message(self):
        self.__add_error_message_general()
        return utils.format_error_message(self.__error_messages)

    def __loginfo(self, msg):
        if self.__print_success_to_console == True:
            print(msg)
        utils.loginfo(LOGGER, msg)

    def __logwarn(self, msg):
        if self.__print_errors_to_console == True:
            print(msg)
        utils.logwarn(LOGGER, msg)



import unittest
import argparse
import logging
import datetime
import uuid
import pika.exceptions
import esgfpid
import esgfpid.utils
from esgfpid.exceptions import MessageNotDeliveredException
from esgfpid.rabbit.exceptions import PIDServerException

# Setup logging:
path = 'logs'
logdate = datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S')
filename = path+'/rabbitmqtest_log_'+logdate+'.log'
esgfpid.utils.ensure_directory_exists(path)
logging.basicConfig(level=logging.DEBUG, filename=filename, filemode='w')
pikalogger = logging.getLogger('pika')
pikalogger.setLevel(logging.INFO)

# Example calls

# Simple call:
#rabbitmqtest -ho foo.com -u user -p passy

# Simple call with a virtual host
#rabbitmqtest -ho foo.com -u user -p passy -vh foo

# Simple call with pika logs to console:
#rabbitmqtest -ho foo.com -u user -p passy -pik
# (This argument can always be added)

# Message test:
#rabbitmqtest -ho foo.com -u user -p passy -m foomessage -ex exch
# (If message given, exchange needs to be given!)

# Production test: Sends a message to production queue 
#rabbitmqtest -ho foo.com -u user -p passy -prod

# -pik (print pika log messages) can always be given
# -px (prefix) can always be passed
# -m (message) needs -ex (exchange). The true exchanges may not be used for this.
# -prod (production test) ignores -m and -exch


# Functions
def close_connection():
    print('Closing connection...')
    if synchronous_mode:
        coupler.done_with_rabbit_business()
    else:
        coupler.finish_rabbit_connection()
    print('Closing connection... done.')

def exit_program():
    print('Exiting...')
    exit()

def send_a_message():
    success = False
    print('\nTesting to send test message to host %s (exchange %s)...' % (host, exch))
    print('Test message: %s' % message)
    try:
        print('Sending message...')
        coupler.send_message_to_queue(message)
        print('Sending message... done.')
        success = True
    except MessageNotDeliveredException as e:
        error_name = e.__class__.__name__
        error_message = str(e)
        print('Sending message... failed. Reason: %s (%s)' % (error_message,error_name))
    except pika.exceptions.ChannelClosed as e:
        error_name = e.__class__.__name__
        error_message = str(e)
        if ('404, "NOT_FOUND - no exchange \'%s\' in vhost' % exch) in str(e):
            print('Sending message... failed. Reason: Wrong exchange name "%s".' % exch)
            print('Exception: %s %s' % (error_name, error_message))
        else:
            print('Sending message... failed. Reason: Channel closed.')
            print('Exception: %s %s' % (error_name, error_message))
    return success


if __name__ == '__main__':
    desc = ('Test the connection to your RabbitMQ host via the esgfpid library.')
    
    # Argument parsing
    parser = argparse.ArgumentParser(description=desc)
    
    # Mandatory, for connection
    parser.add_argument('-ho','--host', nargs=1, action='store', required=True,
                   help='The RabbitMQ host to connect to.')
    parser.add_argument('-u','--user', nargs=1, action='store', required=True,
                   help='The RabbitMQ user name.')
    parser.add_argument('-p','--password', nargs=1, action='store', required=True,
                   help='The RabbitMQ password. Please escape "!" and "$" and "`" with backslashes!')
    # Optional, for connection
    parser.add_argument('-vh', '--virtualhost', nargs='?', action='store',
                    help='The virtual host to connect to.')
    parser.add_argument('-po', '--port', nargs='?', action='store',
                    help='The port to connect to.')
    # Optional, for sending messages:
    parser.add_argument('-m', '--message', nargs='?', action='store',
                    help='The test message to send. Can be any string.')
    parser.add_argument('-ex','--exchange_name', nargs='?', action='store',
                    help='The exchange name to send the message to.')
    parser.add_argument('-px', '--prefix', nargs='?', action='store',
                    help='The handle prefix to use.')
    # Other:
    parser.add_argument('-asy', '--asynchron', action='store_true',
                    help='The test message to send.')
    parser.add_argument('-pik', '--pikalog', action='store_true',
                    help='The test message to send.')
    parser.add_argument('-prod', '--production-test', action='store_true',
                    help='The test message to send.')

    # Parse the args:
    param = parser.parse_args()

    # Get mandatory args
    user = param.user[0]
    password = param.password[0]
    if '\!' in password:
        print('Replacing "\!" in password with "!".')
        password = password.replace('\!', '!')
    host = param.host[0]

    # Get optional args
    message = None
    if param.message is not None:
        message = param.message
    prefix = None
    if param.prefix is not None:
        prefix = param.prefix[0]
    exch = None
    if param.exchange_name is not None:
        exch = param.exchange_name
    prod_test = False
    if param.production_test is not None and param.production_test == True:
        prod_test = True
    synchronous_mode = True
    if param.asynchron is not None and param.asynchron == True:
        synchronous_mode = False
    vhost = None
    if param.virtualhost is not None:
        vhost = param.virtualhost
    port = None
    if param.port is not None:
        port = param.port

    # Switch on pika log:
    if param.pikalog is not None and param.pikalog == True:
        han = logging.StreamHandler()
        formatter = logging.Formatter('\tpika log: %(asctime)s - %(name)s - %(levelname)s - %(message)s')
        han.setFormatter(formatter)
        han.setLevel(logging.INFO)
        pikalogger.addHandler(han)

    # Using CMIP6 prefix as default:
    if prefix is None:
        prefix = esgfpid.defaults.ACCEPTED_PREFIXES[0]

    # Using exchange "foo" as default. If no message is sent, the exchange name does not matter:
    default_exchange = 'foo'
    if exch is None:
        exch = default_exchange

    # Avoid sending message to production exchange (which would confuse the consumer):
    exch_production = 'esgffed-exchange'
    exch_forbidden = [exch_production, 'esgffed-exchange-alternate']
    if exch in exch_forbidden and not prod_test:
        print('Sending a custom message to the exchange "%s" is not allowed.' % exch)
        exit_program()
    
    # Synchronous or asynchronous?
    if not synchronous_mode:
        print('Connecting in asychronous mode...')

    # Should a test message be sent after connection?
    send_message = False
    if message is not None:
        send_message = True

    # Should a test message be sent to production exchange?
    if prod_test:
        suffix = "librarytest"
        handle = '%s/%s' % (prefix, suffix)
        code = str(uuid.uuid1())
        newmessage = '{"operation":"rabbitmqtest", "handle":"%s", "code":"%s"}' % (handle, code)
        print('\nWill test the production exchange (%s).' % exch_production)
        if message is not None:
            print('Ignoring custom message %s, using %s instead.' % (message, newmessage))
        if not exch == default_exchange and not exch == exch_production:
            print('Ignoring exchange %s, using %s instead.' % (exch, exch_production))

        exch = exch_production
        message = newmessage
        send_message = True

   # Make connector
    print('\nTesting connection to RabbitMQ at %s (user %s, password ****)' % (host, user))
    cred = dict(user=user, password=password, url=host)
    if vhost is not None:
        cred['vhost'] = vhost
    if port is not None:
        cred['port'] = port
    connector = esgfpid.Connector(
        handle_prefix=prefix,
        messaging_service_exchange_name=exch,
        messaging_service_credentials=[cred],
        message_service_synchronous=synchronous_mode
    )

    #
    # Test Part 1
    #
    test1_passed = False

    # Connect
    failed = False
    coupler = connector._Connector__coupler
    if synchronous_mode:
        try:
            print('Connecting...')
            coupler.start_rabbit_business()
            test1_passed = True
            print('Connecting... done.')
        except PIDServerException as e:
            error_name = e.__class__.__name__
            error_message = str(e)
            print('Connecting... failed. Reason: %s (%s)' % (error_message,error_name))
            failed = True
    else:
        coupler.start_rabbit_connection()
        print('Parallel thread started...')


    # Did the connection succeed? (Otherwise, no test message)
    if send_message:
        if failed:
            print('\nWill not send test message. Reason: Connection failed.')
            send_message = False
            exit_program()
        elif param.exchange_name is None:
            print('\nWill not send test message. Reason: No exchange name specified. For sending a message, you need to specify an exchange to send to!')
            send_message = False
            close_connection()
            exit_program()

    #
    # Test Part 2
    # Sending a message
    #
    test2_passed = False

    if send_message:
        test2_passed = send_a_message()
    else:
        test2_passed = True

    if send_message and prod_test and test2_passed:
        pass # not implemented
        #print('\nIf the message succeeds, you should find the uuid "%s" somewhere in the record of the handle "%s".' % (code, handle))
        #print('Please checkout:\nhttp://hdl.handle.net/%s' % handle)        
    
    # At the end, close connection
    close_connection()


    # Print final message
    if test1_passed and test2_passed:
        print('\nRabbitMQ connection test done. SUCCESS!')
    else:
        print('\nRabbitMQ connection test done. FAILURE!')



import unittest
import logging
import sys
if sys.version[0] == '2':
    import Queue as queue
else:
    import queue as queue
import pika
import mock
import json
import esgfpid.rabbit.asynchronous.thread_returnhandler
from esgfpid.rabbit.asynchronous.exceptions import OperationNotAllowed

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import *
import resources.TESTVALUES as TESTHELPERS

class ThreadReturnerTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    def make_returnhandler(self, error=None):

        thread = TESTHELPERS.get_thread_mock3(error=error)

        handler = esgfpid.rabbit.asynchronous.thread_returnhandler.UnacceptedMessagesHandler(
            thread)

        return handler, thread

    # Tests

    #
    # Sending messages
    #

    def test_send_message_ok(self):

        # Preparation:
        routing_key = 'foobar'
        emergency_routing_key = esgfpid.utils.RABBIT_EMERGENCY_ROUTING_KEY
        body = '{"foo":"bar", "ROUTING_KEY":"%s"}' % routing_key
        handler, thread = self.make_returnhandler()
        # <Basic.Return(['exchange=rabbitsender_integration_tests', 'reply_code=312', 'reply_text=NO_ROUTE', 'routing_key=cmip6.publisher.HASH.cart.datasets'])>"
        frame = mock.MagicMock()
        frame.reply_text = 'NO_ROUTE'
        frame.routing_key = routing_key
        props = mock.MagicMock()

        # Run code to be tested:
        handler.on_message_not_accepted(thread._channel, frame, props, body)

        # Check result:
        thread.send_a_message.assert_called_once_with(mock.ANY)
        call_args, call_kwargs = thread.send_a_message.call_args
        expected = {"foo": "bar", "ROUTING_KEY": emergency_routing_key, "original_routing_key": routing_key}
        self.assertEqual(json.loads(call_args[0]), expected)

    def test_send_message_second_time(self):

        # Preparation:
        routing_key = 'foobar'
        emergency_routing_key = esgfpid.utils.RABBIT_EMERGENCY_ROUTING_KEY
        body = '{"foo":"bar", "ROUTING_KEY":"%s", "original_routing_key":"%s"}' % (emergency_routing_key, routing_key)
        handler, thread = self.make_returnhandler()
        # <Basic.Return(['exchange=rabbitsender_integration_tests', 'reply_code=312', 'reply_text=NO_ROUTE', 'routing_key=cmip6.publisher.HASH.cart.datasets'])>"
        frame = mock.MagicMock()
        frame.reply_text = 'NO_ROUTE'
        frame.routing_key = emergency_routing_key
        props = mock.MagicMock()

        # Run code to be tested:
        handler.on_message_not_accepted(thread._channel, frame, props, body)

        # Check result:
        thread.send_a_message.assert_not_called()

    def test_send_message_other(self):

        # Preparation:
        routing_key = 'foobar'
        emergency_routing_key = esgfpid.utils.RABBIT_EMERGENCY_ROUTING_KEY
        body = '{"foo":"bar", "ROUTING_KEY":"%s"}' % routing_key
        handler, thread = self.make_returnhandler()
        # <Basic.Return(['exchange=rabbitsender_integration_tests', 'reply_code=312', 'reply_text=NO_ROUTE', 'routing_key=cmip6.publisher.HASH.cart.datasets'])>"
        frame = mock.MagicMock()
        frame.reply_text = 'something'
        frame.routing_key = routing_key
        props = mock.MagicMock()

        # Run code to be tested:
        handler.on_message_not_accepted(thread._channel, frame, props, body)

        # Check result:
        thread.send_a_message.assert_called_once_with(mock.ANY)
        call_args, call_kwargs = thread.send_a_message.call_args
        expected = {"foo": "bar", "ROUTING_KEY": emergency_routing_key, "original_routing_key": routing_key}
        self.assertEqual(json.loads(call_args[0]), expected)


    def test_send_message_no_key(self):

        # Preparation:
        emergency_routing_key = esgfpid.utils.RABBIT_EMERGENCY_ROUTING_KEY
        body = '{"foo":"bar"}'
        handler, thread = self.make_returnhandler()
        # <Basic.Return(['exchange=rabbitsender_integration_tests', 'reply_code=312', 'reply_text=NO_ROUTE', 'routing_key=cmip6.publisher.HASH.cart.datasets'])>"
        frame = mock.MagicMock()
        frame.reply_text = 'NO_ROUTE'
        frame.routing_key = 'whatever'
        props = mock.MagicMock()

        # Run code to be tested:
        handler.on_message_not_accepted(thread._channel, frame, props, body)

        # Check result:
        thread.send_a_message.assert_called_once_with(mock.ANY)
        call_args, call_kwargs = thread.send_a_message.call_args
        expected = {"foo": "bar", "ROUTING_KEY": emergency_routing_key, "original_routing_key": "None"}
        self.assertEqual(json.loads(call_args[0]), expected)


    def test_send_message_error(self):

        # Preparation:
        emergency_routing_key = esgfpid.utils.RABBIT_EMERGENCY_ROUTING_KEY
        body = '{"foo":"bar"}'
        handler, thread = self.make_returnhandler(error=pika.exceptions.ChannelClosed(reply_code=0, reply_text='Channel Closed'))
        # <Basic.Return(['exchange=rabbitsender_integration_tests', 'reply_code=312', 'reply_text=NO_ROUTE', 'routing_key=cmip6.publisher.HASH.cart.datasets'])>"
        frame = mock.MagicMock()
        frame.reply_text = 'NO_ROUTE'
        frame.routing_key = 'whatever'
        props = mock.MagicMock()

        # Run code to be tested:
        handler.on_message_not_accepted(thread._channel, frame, props, body)

        # Check result:
        # Can not check if error was raised...
        thread.send_a_message.assert_called_once_with(mock.ANY)
        call_args, call_kwargs = thread.send_a_message.call_args
        expected = {"foo": "bar", "ROUTING_KEY": emergency_routing_key, "original_routing_key": "None"}
        self.assertEqual(json.loads(call_args[0]), expected)



import esgfpid
import tests.mocks.solrmock
import tests.mocks.rabbitmock

TESTVALUES = dict(
    solr_url='http://solr.foo',
    prefix='21.14foo',
    drs_id1='my/drs/id',
    version_number1=201603,
    suffix1='afd65cd0-9296-35bc-a706-be98665c9c36',
    data_node='esgf-foo.dkrz.de',
    url_messaging_service='http://rabbit.foo',
    messaging_exchange='exch_name',
    exchange_name='exch', # TODO JUST ONE NAME!
    thredds_service_path='test/thr/path',
    rabbit_username='RogerRabbit',
    rabbit_password='carrotDreams',
    errata_ids=['123456','654321'],
    errata_id='123456',
    file_name1='mytestfilename.nc',
    file_suffix1='xyz123abc',
    file_suffix2='xyz123def',
    checksum1='testchecksumabc',
    file_size1=10010,
    publish_path1='my/publish/path',
    checksum_type1='SHA999',
    file_version1='987'
)


# Test variables:
#testcoupler = self.__make_patched_testcoupler(solr_off=True)
solr_off = True
testcoupler = esgfpid.coupling.Coupler(
    handle_prefix = TESTVALUES['prefix'],
    messaging_service_urls = 'rabbit_should_not_be_used',
    messaging_service_url_preferred = None,
    messaging_service_exchange_name = TESTVALUES['messaging_exchange'],
    messaging_service_username = TESTVALUES['rabbit_username'],
    messaging_service_password = TESTVALUES['rabbit_password'],
    data_node = TESTVALUES['data_node'],
    thredds_service_path = TESTVALUES['thredds_service_path'],
    solr_url = 'solr_should_not_be_used',
    solr_https_verify=True,
    solr_switched_off=solr_off
)
# Replace objects that interact with servers with mocks
# Patch rabbit
rabbitmock = tests.mocks.rabbitmock.SimpleMockRabbitSender()
testcoupler._Coupler__rabbit_message_sender = rabbitmock



# Step 1: Make assistant for dataset
datasetargs = dict(
    prefix=TESTVALUES['prefix'],
    drs_id=TESTVALUES['drs_id1'],
    version_number=TESTVALUES['version_number1'],
    is_replica=False,
    thredds_service_path=TESTVALUES['thredds_service_path'],
    data_node=TESTVALUES['data_node'],
    coupler=testcoupler,
    consumer_solr_url='does_not_matter'
)
assistant = esgfpid.assistant.publish.DatasetPublicationAssistant(**datasetargs)

# Step 2: Add files
fileargs =  dict(
    file_name = TESTVALUES['file_name1'],
    file_handle = 'hdl:'+TESTVALUES['prefix']+'/'+TESTVALUES['file_suffix1'],
    checksum = TESTVALUES['checksum1'],
    file_size = TESTVALUES['file_size1'],
    publish_path = TESTVALUES['publish_path1'],
    prefix = TESTVALUES['prefix'],
    checksum_type=TESTVALUES['checksum_type1'],
    file_version=TESTVALUES['file_version1'] # string
)
#assistant.add_file(**fileargs)

# Step 3: Finish publication
assistant.dataset_publication_finished()


import esgfpid.utils
from esgfpid.utils import ROUTING_KEYS

'''
The messages module creates the JSON messages to be sent to the rabbit.
It does not check the type of the values passed. This has to be checked in the assistants,
otherwise it is redundant.
'''

JSON_KEY_ROUTING_KEY = 'ROUTING_KEY'


def publish_file(**args):

    # Check args:
    mandatory_args = ['file_handle', 'is_replica', 'file_size', 'file_name', 'checksum',
                      'data_url', 'parent_dataset', 'timestamp', 'checksum_type',
                      'file_version', 'data_node']
    esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)

    # Message:
    message = dict(
        handle = args['file_handle'],
        aggregation_level = 'file',
        operation = 'publish',
        is_replica=args['is_replica'],
        file_name=args['file_name'],
        file_size=args['file_size'],
        checksum=args['checksum'],
        data_url=args['data_url'],
        data_node=args['data_node'],
        parent_dataset=args['parent_dataset'],
        message_timestamp = args['timestamp'],
        checksum_type = args['checksum_type'],
        file_version = args['file_version'] # can be int or string or ...
    )

    # Routing key:
    routing_key = ROUTING_KEYS['publi_file']
    if args['is_replica'] == True: # Publish Assistant parses this to boolean!
        routing_key = ROUTING_KEYS['publi_file_rep']
    message[JSON_KEY_ROUTING_KEY] = routing_key
    return message

def publish_dataset(**args):

    # Check args:
    mandatory_args = ['dataset_handle', 'drs_id', 'is_replica', 'version_number',
                      'list_of_files', 'data_node', 'timestamp']
    esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)

    # Message:
    message = dict(
        handle = args['dataset_handle'],
        aggregation_level = 'dataset',
        operation = 'publish',
        drs_id = args['drs_id'],
        is_replica=args['is_replica'],
        version_number=args['version_number'], # Publish Assistant parses this to int!
        files=args['list_of_files'],
        data_node=args['data_node'],
        message_timestamp = args['timestamp']
    )

    # Optional:
    if 'consumer_solr_url' in args and args['consumer_solr_url'] is not None:
        message['consumer_solr_url'] = args['consumer_solr_url']

    # Routing key:
    routing_key = ROUTING_KEYS['publi_ds']
    if args['is_replica'] == True: # Publish Assistant parses this to boolean!
        routing_key = ROUTING_KEYS['publi_ds_rep']
    message[JSON_KEY_ROUTING_KEY] = routing_key

    return message

def unpublish_allversions_consumer_must_find_versions(**args):
    mandatory_args = ['drs_id', 'data_node', 'timestamp']
    esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
    message = dict(
        operation = 'unpublish_all_versions',
        aggregation_level = 'dataset',
        message_timestamp = args['timestamp'],
        drs_id = args['drs_id'],
        data_node=args['data_node']
    )
    # Optional:
    if 'consumer_solr_url' in args and args['consumer_solr_url'] is not None:
        message['consumer_solr_url'] = args['consumer_solr_url']
    routing_key = ROUTING_KEYS['unpubli_all']
    message[JSON_KEY_ROUTING_KEY] = routing_key
    return message


def unpublish_one_version(**args):
    mandatory_args = ['data_node', 'timestamp', 'dataset_handle', 'drs_id']
    esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
    message = dict(
        operation = 'unpublish_one_version',
        aggregation_level = 'dataset',
        message_timestamp = args['timestamp'],
        handle = args['dataset_handle'],
        data_node=args['data_node'],
        drs_id = args['drs_id']
    )
    if 'version_number' in args:
        message['version_number'] = args['version_number']
    routing_key = ROUTING_KEYS['unpubli_one']
    message[JSON_KEY_ROUTING_KEY] = routing_key
    return message

def add_errata_ids_message(**args):
    mandatory_args = ['dataset_handle', 'timestamp', 'errata_ids', 'drs_id', 'version_number']
    esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
    message = dict(
        handle = args['dataset_handle'],
        message_timestamp = args['timestamp'],
        errata_ids = args['errata_ids'],
        operation = 'add_errata_ids',
        drs_id = args['drs_id'],
        version_number = args['version_number']
    )
    routing_key = ROUTING_KEYS['err_add']
    message[JSON_KEY_ROUTING_KEY] = routing_key
    return message

def remove_errata_ids_message(**args):
    mandatory_args = ['dataset_handle', 'timestamp', 'errata_ids', 'drs_id', 'version_number']
    esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
    message = dict(
        handle = args['dataset_handle'],
        message_timestamp = args['timestamp'],
        errata_ids = args['errata_ids'],
        operation = 'remove_errata_ids',
        drs_id = args['drs_id'],
        version_number = args['version_number']
    )    
    routing_key = ROUTING_KEYS['err_rem']
    message[JSON_KEY_ROUTING_KEY] = routing_key

    return message

def make_data_cart_message(**args):
    mandatory_args = ['cart_handle', 'timestamp', 'data_cart_content']
    esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
    message = dict(
        handle = args['cart_handle'],
        message_timestamp = args['timestamp'],
        data_cart_content = args['data_cart_content'],
        operation = 'shopping_cart'
    )
    routing_key = ROUTING_KEYS['data_cart']
    message[JSON_KEY_ROUTING_KEY] = routing_key
    return message


import logging
import esgfpid.exceptions
import esgfpid.assistant.consistency
import esgfpid.assistant.messages
import esgfpid.utils as utils
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Helper:

def create_dataset_handle(**args):
    return utils.make_handle_from_drsid_and_versionnumber(
        drs_id=args['drs_id'],
        version_number=args['version_number'],
        prefix=args['prefix']
    )

class DatasetPublicationAssistant(object):

    def __init__(self, **args):
        logdebug(LOGGER, 'Constructor for Publication assistant for dataset "%s", version "%s" at host "%s".',
            args['drs_id'],
            args['version_number'],
            args['data_node']
        )

        # Check args
        mandatory_args = ['drs_id', 'version_number', 'data_node', 'prefix',
                          'thredds_service_path', 'is_replica', 'coupler',
                          'consumer_solr_url']
        optional_args = []
        utils.check_presence_of_mandatory_args(args, mandatory_args)
        utils.add_missing_optional_args_with_value_none(args, optional_args)
        self.__enforce_integer_version_number(args)
        self.__enforce_boolean_replica_flag(args)

        # Init methods...
        self.__store_args_in_attributes(args)
        self.__define_other_attributes()
        self.__create_and_store_dataset_handle()
        self.__init_state_machine()

        logdebug(LOGGER, 'Done: Constructor for Publication assistant for dataset "%s", version "%i" at host "%s".',
            args['drs_id'],
            args['version_number'],
            args['data_node']
        )

    def __enforce_integer_version_number(self, args):
        try:
            args['version_number'] = int(args['version_number'])
        except ValueError:
            raise esgfpid.exceptions.ArgumentError('Dataset version number is not an integer')

    def __enforce_boolean_replica_flag(self, args):
        try:
            args['is_replica'] = utils.get_boolean(args['is_replica'])
        except ValueError:
            msg = ('Replica flag "%s" could not be parsed to boolean. '
                   'Please pass a boolean or "True" or "true" or "False" or "false"'
                   % args['is_replica'])
            raise esgfpid.exceptions.ArgumentError(msg)

    def __enforce_integer_file_size(self, args):
        try:
            args['file_size'] = int(args['file_size'])
        except ValueError:
            raise esgfpid.exceptions.ArgumentError('File size is not an integer')

    def __enforce_string_file_version(self, args):
        args['file_version'] = str(args['file_version'])

    def __define_other_attributes(self):
        self.__dataset_handle = None
        self.__list_of_file_handles = []
        self.__list_of_file_messages = []
        self.__message_timestamp = utils.get_now_utc_as_formatted_string()

    def __store_args_in_attributes(self, args):
        self.__drs_id = args['drs_id']
        self.__version_number = args['version_number']
        self.__data_node = args['data_node'].rstrip('/')
        self.__prefix = args['prefix']
        self.__thredds_service_path = args['thredds_service_path'].strip('/')
        self.__is_replica = args['is_replica']
        self.__coupler = args['coupler']
        self.__consumer_solr_url = args['consumer_solr_url']

    def __init_state_machine(self):
        self.__machine_states = {'dataset_added':0, 'files_added':1, 'publication_finished':2}
        self.__machine_state = self.__machine_states['dataset_added']

    def __create_and_store_dataset_handle(self):
        self.__dataset_handle = create_dataset_handle(
            drs_id = self.__drs_id,
            version_number = self.__version_number,
            prefix = self.__prefix
        )

    def get_dataset_handle(self):
        '''
        This returns the handle string of the dataset to be
        published, so that the publisher can use it for its
        own purposes, e.g. publishing it on a website.

        The handle string consists of the prefix specified
        at library init, and a suffix. The suffix is created
        by making a hash over dataset id and version number.

        :return: The handle string of this dataset,
            e.g.: "hdl:21.14100/foobar".
        '''
        return self.__dataset_handle

    # work horses:

    def add_file(self, **args):
        '''
        Adds a file's information to the set of files to be
        published in this dataset.

        :param file_name: Mandatory. The file name (string).
            This information will simply be included in the
            PID record, but not used for anything.

        :param file_handle: Mandatory. The handle (PID) of
            this file (string). It is included in the file's netcdf
            header. It must bear the prefix that this library 
            (or rather, the consuming servlet that will consume
            this library's requests), has write access to.

        :param file_size: Mandatory. The file size (as string or
            integer. Will be transformed to integer). This
            information will be included in the handle record
            and used for consistency checks during republications
            of files with the same handle.

        :param checksum: Mandatory. The file's checksum. This
            information will be included in the handle record
            and used for consistency checks during republications
            of files with the same handle.

        :param checksum_type: Mandatory. The checksum type/method
            (string), e.g. "MD5" or "SHA256". This information will
            be included in the handle record and used for consistency
            checks during republications of files with the same handle.

        :param publish_path: Mandatory. The THREDDS publish path as
            a string. This is part of the URL for accessing the file,
            which will be part of the handle record. It will not be
            accessed, neither by the library nor by the consumer.
            The URL consists of the dataset's "data_node", the dataset's 
            "thredds_service_path", and this "publish_path". Redundant
            slashes are removed. If the URL does not start with "http",
            "http://" is added.

        :param file_version: Mandatory. Any string. File versions
            are not managed in the PID. This information will simply be
            included in the PID record, but not used for any reasoning.
        '''

        # Check if allowed:
        self.__check_if_adding_files_allowed_right_now()

        # Check if args ok:
        mandatory_args = ['file_name', 'file_handle', 'file_size',
                          'checksum', 'publish_path', 'checksum_type',
                          'file_version']
        utils.check_presence_of_mandatory_args(args, mandatory_args)
        self.__enforce_integer_file_size(args)
        self.__enforce_string_file_version(args)

        # Add file:
        self.__check_and_correct_handle_syntax(args)
        self.__add_file(**args)

    def __check_and_correct_handle_syntax(self, args):
        self.__make_sure_hdl_is_added(args)
        self.__check_if_prefix_is_there(args['file_handle'])

    def __add_file(self, **args):
        logdebug(LOGGER, 'Adding file "%s" with handle "%s".', args['file_name'], args['file_handle'])
        self.__add_file_to_datasets_children(args['file_handle'])
        self.__adapt_file_args(args)
        self.__create_and_store_file_publication_message(args)        
        self.__set_machine_state_to_files_added()
        logtrace(LOGGER, 'Adding file done.')

    def __add_file_to_datasets_children(self, file_handle):
        self.__list_of_file_handles.append(file_handle)

    def __make_sure_hdl_is_added(self, args):
        if not args['file_handle'].startswith('hdl:'):
            args['file_handle'] = 'hdl:'+args['file_handle']

    def __check_if_prefix_is_there(self, file_handle):
        if not file_handle.startswith('hdl:'+self.__prefix+'/'):

            expected = self.__prefix + '/'+ file_handle.lstrip('hdl:')
            msg = ('\nThis file\'s tracking_id "%s" does not have the expected handle prefix "%s".'
                   '\nExpected "%s" or "%s"' % (file_handle, self.__prefix, expected, 'hdl:'+expected))

            if '/' in file_handle:
                maybe_prefix = file_handle.split('/')[0].lstrip('hdl:')
                msg += ('.\nIf "%s" was meant to be the prefix, it does not correspond to '
                        'the prefix specified when initializing this library' % maybe_prefix)

            raise esgfpid.exceptions.ESGFException(msg)


    def __adapt_file_args(self, args):
        url = self.__create_file_url(args['publish_path'])
        args['data_url'] = url
        del args['publish_path']

    def __create_file_url(self, publish_path):
        url = self.__data_node +'/'+ self.__thredds_service_path +'/'+ publish_path.strip('/')
        if not url.startswith('http'):
            url = 'http://'+url
        return url

    def __create_and_store_file_publication_message(self, args):
        message = self.__create_file_publication_message(args)
        self.__list_of_file_messages.append(message)

    def __set_machine_state_to_files_added(self):
        self.__machine_state = self.__machine_states['files_added']

    def __check_if_adding_files_allowed_right_now(self):
        dataset_added = self.__machine_state == self.__machine_states['dataset_added']
        files_added = self.__machine_state == self.__machine_states['files_added']
        if dataset_added or files_added:
            pass
        else:
            msg = 'Too late to add files!'
            logwarn(LOGGER, msg)
            raise esgfpid.exceptions.OperationUnsupportedException(msg)

    def dataset_publication_finished(self, ignore_exception=False):
        '''
        This is the "commit". It triggers the creation/update of handles.
        
        * Check if the set of files corresponds to the previously published set (if applicable, and if solr url given, and if solr replied)
        * The dataset publication message is created and sent to the queue.
        * All file publication messages are sent to the queue.

        '''
        self.__check_if_dataset_publication_allowed_right_now()
        self.__check_data_consistency(ignore_exception)
        self.__coupler.start_rabbit_business() # Synchronous: Opens connection. Asynchronous: Ignored.
        self.__create_and_send_dataset_publication_message_to_queue()
        self.__send_existing_file_messages_to_queue()
        self.__coupler.done_with_rabbit_business() # Synchronous: Closes connection. Asynchronous: Ignored.
        self.__set_machine_state_to_finished()
        loginfo(LOGGER, 'Requesting to publish PID for dataset "%s" (version %s) and its files at "%s" (handle %s).', self.__drs_id, self.__version_number, self.__data_node, self.__dataset_handle)

    def __check_if_dataset_publication_allowed_right_now(self):
        if not self.__machine_state == self.__machine_states['files_added']:
            msg = None
        
            if self.__machine_state == self.__machine_states['dataset_added']:
                msg = 'No file added yet.'
            else:
                msg = 'Publication was already done.'

            logwarn(LOGGER, msg)
            raise esgfpid.exceptions.OperationUnsupportedException(msg)


    def __check_data_consistency(self, ignore_exception):
        checker = esgfpid.assistant.consistency.Checker(
            coupler=self.__coupler,
            drs_id=self.__drs_id,
            version_number=self.__version_number,
            data_node=self.__data_node
        )
        check_possible = checker.can_run_check()
        if check_possible:
            check_passed = checker.data_consistency_check(self.__list_of_file_handles)
            if check_passed:
                loginfo(LOGGER, 'Data consistency check passed for dataset %s.', self.__dataset_handle)
            else:
                msg = 'Dataset consistency check failed'
                logwarn(LOGGER, msg)
                if not ignore_exception:
                    raise esgfpid.exceptions.InconsistentFilesetException(msg)
        else:
            logdebug(LOGGER, 'No consistency check was carried out.')

    def __create_and_send_dataset_publication_message_to_queue(self):
        self.__remove_duplicates_from_list_of_file_handles()
        message = self.__create_dataset_publication_message()
        self.__send_message_to_queue(message)
        logdebug(LOGGER, 'Dataset publication message handed to rabbit thread.')
        logtrace(LOGGER, 'Dataset publication message: %s (%s, version %s).', self.__dataset_handle, self.__drs_id, self.__version_number)

    def __remove_duplicates_from_list_of_file_handles(self):
        self.__list_of_file_handles = list(set(self.__list_of_file_handles))

    def __send_existing_file_messages_to_queue(self):
        for i in range(0, len(self.__list_of_file_messages)):
            self.__try_to_send_one_file_message(i)
        msg = 'All file publication jobs handed to rabbit thread.'
        logdebug(LOGGER, msg)
        
    def __try_to_send_one_file_message(self, list_index):
        msg = self.__list_of_file_messages[list_index]
        success = self.__send_message_to_queue(msg)
        logdebug(LOGGER, 'File publication message handed to rabbit thread: %s (%s)', msg['handle'], msg['file_name'])
        return success

    def __set_machine_state_to_finished(self):
        self.__machine_state = self.__machine_states['publication_finished']

    def __create_file_publication_message(self, args):
        
        message = esgfpid.assistant.messages.publish_file(
            file_handle=args['file_handle'],
            file_size=args['file_size'],
            file_name=args['file_name'],
            checksum=args['checksum'],
            data_url=args['data_url'],
            data_node=self.__data_node,
            parent_dataset=self.__dataset_handle,
            checksum_type=args['checksum_type'],
            file_version=args['file_version'],
            is_replica=self.__is_replica,
            timestamp=self.__message_timestamp,
        )
        return message

    def __create_dataset_publication_message(self):

        message = esgfpid.assistant.messages.publish_dataset(
            dataset_handle=self.__dataset_handle,
            is_replica=self.__is_replica,
            drs_id=self.__drs_id,
            version_number=self.__version_number,
            list_of_files=self.__list_of_file_handles,
            data_node=self.__data_node,
            timestamp=self.__message_timestamp,
            consumer_solr_url=self.__consumer_solr_url
        )
        return message

    def __send_message_to_queue(self, message):
        success = self.__coupler.send_message_to_queue(message)
        return success


import unittest
import mock
import logging
import datetime
import time
import esgfpid.rabbit.asynchronous
from esgfpid.rabbit.asynchronous.exceptions import OperationNotAllowed

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import *
import resources.TESTVALUES as TESTHELPERS

import globalvar
if globalvar.QUICK_ONLY:
    print('Skipping slow tests in module "%s".' % __name__)


class RabbitAsynConnectorTestCase(unittest.TestCase):

    slow_message = '\nRunning a slow test (avoid by using -ls flag).'


    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    def assert_messages_are_in_queue(self, msg_queue, list_of_messages):
        queue_content = []
        while not msg_queue.empty():
            queue_content.append(msg_queue.get(False))
        for msg in list_of_messages:
            self.assertIn(msg, queue_content)

    #
    # Init
    #

    '''
    Test whether instances of rabbitconnector and thread
    are created.
    '''
    def test_init_ok(self):

        # Test variables:
        nodemanager = TESTHELPERS.get_nodemanager()

        # Run code to be tested:
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(nodemanager)

        # Check result:
        self.assertIsInstance(testrabbit, esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector, 'Constructor fail.')
        self.assertTrue(testrabbit._AsynchronousRabbitConnector__not_started_yet)
        thread = testrabbit._AsynchronousRabbitConnector__thread
        self.assertIsInstance(thread, esgfpid.rabbit.asynchronous.rabbitthread.RabbitThread, 'Constructor fail.')


    #
    # Start thread
    #

    '''
    Test whether the start method does the necessary things.
    We expect it to change the state, and to run the thread.
    '''
    @unittest.skipIf(globalvar.QUICK_ONLY, '(this test is slow)')
    def test_start_thread_ok(self):

        print(self.slow_message)

        # Preparation:
        nodemanager = TESTHELPERS.get_nodemanager()
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(nodemanager)
        
        # Mock the builder (so it cannot start a connection):
        def side_effect_first_connection():
            LOGGER.debug('Pretending to do something in the thread.run()')
            time.sleep(0.5)
            LOGGER.debug('Finished pretending to do something in the thread.run()')
        buildermock = mock.MagicMock()
        buildermock.first_connection = mock.MagicMock()
        buildermock.first_connection.side_effect = side_effect_first_connection
        testrabbit._AsynchronousRabbitConnector__thread._RabbitThread__builder = buildermock

        # Check preconditions:
        self.assertTrue(testrabbit._AsynchronousRabbitConnector__not_started_yet)
        self.assertTrue(testrabbit._AsynchronousRabbitConnector__statemachine.is_NOT_STARTED_YET())

        # Run code to be tested:
        # This runs the thread, which triggers building a connection.
        # In this test, it calls the side effect defined above, which blocks for
        # a second. Afterwards, the thread should be finished.
        testrabbit.start_rabbit_thread()

        # Check results:
        # Check if thread is alive:
        self.assertTrue(testrabbit._AsynchronousRabbitConnector__thread.is_alive())
        # Join the thread...
        print("Joining...")
        testrabbit._AsynchronousRabbitConnector__thread.join()
        print("Joining done...")
        # Check if the thread has ended:
        self.assertFalse(testrabbit._AsynchronousRabbitConnector__thread.is_alive())
        # Check state:
        self.assertFalse(testrabbit._AsynchronousRabbitConnector__not_started_yet)
        self.assertTrue(testrabbit._AsynchronousRabbitConnector__statemachine.is_WAITING_TO_BE_AVAILABLE())
        # Check if run was called:
        buildermock.first_connection.assert_called()


    #
    # Sending messages
    #

    '''
    Test behaviour when we try sending messages but the
    thread was not started yet.
    It should raise an exception.
    '''
    def test_send_message_not_started_yet(self):

        # Preparation:
        nodemanager = TESTHELPERS.get_nodemanager()
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(nodemanager)
        self.assertTrue(testrabbit._AsynchronousRabbitConnector__not_started_yet)

        # Run code to be tested:
        with self.assertRaises(OperationNotAllowed):
            testrabbit.send_message_to_queue('message-foo')
        with self.assertRaises(OperationNotAllowed):
            testrabbit.send_many_messages_to_queue(['a','b','c'])

    '''
    Test behaviour when we try sending messages but the
    thread was not started yet.
    It should raise an exception.
    '''
    def test_send_message_not_started_yet_2(self):

        # Preparation:
        nodemanager = TESTHELPERS.get_nodemanager()
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(nodemanager)
        testrabbit._AsynchronousRabbitConnector__not_started_yet = False
        self.assertFalse(testrabbit._AsynchronousRabbitConnector__not_started_yet)
        self.assertTrue(testrabbit._AsynchronousRabbitConnector__statemachine.is_NOT_STARTED_YET())

        # Run code to be tested:
        with self.assertRaises(OperationNotAllowed):
            testrabbit.send_message_to_queue('message-foo')
        with self.assertRaises(OperationNotAllowed):
            testrabbit.send_many_messages_to_queue(['a','b','c'])

    '''
    Test behaviour when we send messages when the thread
    was properly started.

    We expect the message to be put into the queue.
    We expect the publish event to be handed by the connection
    to the feeder module.
    '''
    def test_send_message_ok(self):

        # Preparations
        nodemanager = TESTHELPERS.get_nodemanager()
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(nodemanager)
        testrabbit._AsynchronousRabbitConnector__statemachine.set_to_available()
        testrabbit._AsynchronousRabbitConnector__not_started_yet = False

        # Mock the connection (it has to hand the event over to the feeder mock):
        connectionmock = testrabbit._AsynchronousRabbitConnector__thread._connection = TESTHELPERS.get_connection_mock()

        # Mock the feeder (it has to receive the publish event):
        feedermock = testrabbit._AsynchronousRabbitConnector__thread._RabbitThread__feeder = mock.MagicMock()

        # Run code to be tested:
        testrabbit.send_message_to_queue('foo')
        testrabbit.send_many_messages_to_queue(['a','b','c'])

        # Check that publish was called:
        feedermock.publish_message.assert_called()
        self.assertTrue(feedermock.publish_message.call_count>=4)

        # Check that the four messages were put into the queue:
        msg_queue = testrabbit._AsynchronousRabbitConnector__unpublished_messages_queue
        self.assert_messages_are_in_queue(msg_queue, ['foo', 'a', 'b', 'c'])


    def test_send_message_waiting(self):

        # Preparations
        nodemanager = TESTHELPERS.get_nodemanager()
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(nodemanager)
        testrabbit._AsynchronousRabbitConnector__statemachine.set_to_waiting_to_be_available()
        testrabbit._AsynchronousRabbitConnector__not_started_yet = False

        # Mock the connection (it has to hand the event over to the feeder mock):
        connectionmock = testrabbit._AsynchronousRabbitConnector__thread._connection = TESTHELPERS.get_connection_mock()

        # Mock the feeder (it has to receive the publish event):
        feedermock = testrabbit._AsynchronousRabbitConnector__thread._RabbitThread__feeder = mock.MagicMock()

        # Run code to be tested:
        testrabbit.send_many_messages_to_queue(['a','b','c'])
        testrabbit.send_message_to_queue('foo')

        # Check that publish was NOT called:
        feedermock.publish_message.assert_not_called()

        # Check that the four messages were put into the queue:
        msg_queue = testrabbit._AsynchronousRabbitConnector__unpublished_messages_queue
        self.assert_messages_are_in_queue(msg_queue, ['foo', 'a', 'b', 'c'])

    def test_send_message_unavail(self):

        # Preparations
        nodemanager = TESTHELPERS.get_nodemanager()
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(nodemanager)
        testrabbit._AsynchronousRabbitConnector__statemachine.set_to_permanently_unavailable()
        testrabbit._AsynchronousRabbitConnector__not_started_yet = False

        # Mock the feeder (it has to receive the publish event):
        feedermock = testrabbit._AsynchronousRabbitConnector__thread._RabbitThread__feeder = mock.MagicMock()

        # Mock the connection (it has to hand the event over to the feeder mock):
        connectionmock = testrabbit._AsynchronousRabbitConnector__thread._connection = TESTHELPERS.get_connection_mock()

        # Run code to be tested:
        testrabbit.send_message_to_queue('foo')
        testrabbit.send_many_messages_to_queue(['a','b','c'])

        # Check that publish was NOT called:
        feedermock.publish_message.assert_not_called()

        # Check that the four messages were NOT put into the queue:
        msg_queue = testrabbit._AsynchronousRabbitConnector__unpublished_messages_queue
        self.assertTrue(msg_queue.empty())

    def test_send_message_user_closed(self):

        # Preparations
        nodemanager = TESTHELPERS.get_nodemanager()
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(nodemanager)
        testrabbit._AsynchronousRabbitConnector__statemachine.set_to_permanently_unavailable()
        testrabbit._AsynchronousRabbitConnector__statemachine.set_detail_closed_by_publisher()
        testrabbit._AsynchronousRabbitConnector__not_started_yet = False

        # Mock the feeder (it has to receive the publish event):
        feedermock = testrabbit._AsynchronousRabbitConnector__thread._RabbitThread__feeder = mock.MagicMock()

        # Mock the connection (it has to hand the event over to the feeder mock):
        connectionmock = testrabbit._AsynchronousRabbitConnector__thread._connection = TESTHELPERS.get_connection_mock()

        # Run code to be tested:
        with self.assertRaises(OperationNotAllowed):
            testrabbit.send_message_to_queue('foo')
        with self.assertRaises(OperationNotAllowed):
            testrabbit.send_many_messages_to_queue(['a','b','c'])

        # Check that publish was NOT called:
        feedermock.publish_message.assert_not_called()

        # Check that the four messages were NOT put into the queue:
        msg_queue = testrabbit._AsynchronousRabbitConnector__unpublished_messages_queue
        self.assertTrue(msg_queue.empty())


    #
    # Gently finish
    #

    def test_gently_finish_ok(self):

        # Preparations
        nodemanager = TESTHELPERS.get_nodemanager()
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(nodemanager)
        testrabbit._AsynchronousRabbitConnector__statemachine.set_to_available()

        # Mock the wait-event, otherwise the library blocks, because the
        # wait event would be unblocked in a patched function...
        testrabbit._AsynchronousRabbitConnector__thread._RabbitThread__gently_finish_ready = wait_event_mock = mock.MagicMock()

        # Mock the join function, as it cannot join if the thread was not started.
        testrabbit._AsynchronousRabbitConnector__thread.join =  joinmock = mock.MagicMock()

        # Mock shutter (it has to receive the close event)
        testrabbit._AsynchronousRabbitConnector__thread._RabbitThread__shutter = shuttermock = mock.MagicMock()

        # Mock the connection (it has to hand the event over to the feeder mock):
        connectionmock = testrabbit._AsynchronousRabbitConnector__thread._connection = TESTHELPERS.get_connection_mock()

        # Run code to be tested:
        testrabbit.finish_rabbit_thread()

        # Check result
        shuttermock.finish_gently.assert_called()
        wait_event_mock.wait.assert_called()
        joinmock.assert_called()

    #
    # Force finish
    #

    def test_force_finish_ok(self):

        # Preparations
        nodemanager = TESTHELPERS.get_nodemanager()
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(nodemanager)
        testrabbit._AsynchronousRabbitConnector__statemachine.set_to_available()

        # Mock the join function, as it cannot join if the thread was not started.
        testrabbit._AsynchronousRabbitConnector__thread.join =  joinmock = mock.MagicMock()

        # Mock shutter (it has to receive the close event)
        testrabbit._AsynchronousRabbitConnector__thread._RabbitThread__shutter = shuttermock = mock.MagicMock()
        
        # Mock the connection (it has to hand the event over to the feeder mock):
        connectionmock = testrabbit._AsynchronousRabbitConnector__thread._connection = TESTHELPERS.get_connection_mock()

        # Run code to be tested:
        testrabbit.force_finish_rabbit_thread()

        # Check result
        shuttermock.force_finish.assert_called()
        joinmock.assert_called()

    @unittest.skipIf(globalvar.QUICK_ONLY, '(this test is slow)')
    def test_force_finish_join_fails(self):

        print(self.slow_message)

        # Preparations
        nodemanager = TESTHELPERS.get_nodemanager()
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(nodemanager)
        testrabbit._AsynchronousRabbitConnector__statemachine.set_to_available()

        # Mock the join function, as it cannot join if the thread was not started.
        testrabbit._AsynchronousRabbitConnector__thread.join =  joinmock = mock.MagicMock()
        testrabbit._AsynchronousRabbitConnector__thread.is_alive =  alivemock = mock.MagicMock()
        alivemock.side_effect = [True,False,False,False,False,False,False]

        # Mock shutter (it has to receive the close event)
        testrabbit._AsynchronousRabbitConnector__thread._RabbitThread__shutter = shuttermock = mock.MagicMock()
        
        # Mock the connection (it has to hand the event over to the feeder mock):
        connectionmock = testrabbit._AsynchronousRabbitConnector__thread._connection = TESTHELPERS.get_connection_mock()

        # Run code to be tested:
        testrabbit.force_finish_rabbit_thread()

        # Check result
        shuttermock.force_finish.assert_called()
        joinmock.assert_called()


    #
    # Kleinkram
    #

    def test_is_finished(self):

        # Preparation:
        nodemanager = TESTHELPERS.get_nodemanager()
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(nodemanager)

        # Mock thread.is_alive and test getter for it:
        thread = testrabbit._AsynchronousRabbitConnector__thread
        thread.is_alive = mock.MagicMock()
        thread.is_alive.return_value = True
        self.assertFalse(testrabbit.is_finished())
        thread.is_alive.return_value = False
        self.assertTrue(testrabbit.is_finished())

    def test_tell_stop_waiting(self):

        # Preparation:
        nodemanager = TESTHELPERS.get_nodemanager()
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(nodemanager)
        testrabbit._AsynchronousRabbitConnector__thread._RabbitThread__gently_finish_ready = mock.MagicMock()

        # Run code to be tested:
        testrabbit._AsynchronousRabbitConnector__thread.tell_publisher_to_stop_waiting_for_gentle_finish()

        # Check result
        testrabbit._AsynchronousRabbitConnector__thread._RabbitThread__gently_finish_ready.set.assert_called()

    def test_unblock_events(self):

        # Preparation:
        nodemanager = TESTHELPERS.get_nodemanager()
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(nodemanager)

        # Run code to be tested:
        testrabbit._AsynchronousRabbitConnector__thread.unblock_events()

        # Can't really check results... Cannot block the events to unblock them here,
        # without opening a thread...

    def test_wait_for_connection(self):

        # Preparation:
        nodemanager = TESTHELPERS.get_nodemanager()
        testrabbit = esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(nodemanager)
        testrabbit._AsynchronousRabbitConnector__statemachine.set_to_available()
        testrabbit._AsynchronousRabbitConnector__not_started_yet = False

        # Mock the feeder (it has to receive the publish event):
        feedermock = testrabbit._AsynchronousRabbitConnector__thread._RabbitThread__feeder = mock.MagicMock()

        # Mock event...
        # So when the synchronisation is done, a connection is there...
        # When the event is called, we mock the connection (it has to hand the event over to the feeder mock):

        def sync_side_effect():
            connectionmock = testrabbit._AsynchronousRabbitConnector__thread._connection = TESTHELPERS.get_connection_mock()
        sync_event_mock = mock.MagicMock()
        sync_event_mock.wait.side_effect = sync_side_effect
        testrabbit._AsynchronousRabbitConnector__thread._RabbitThread__connection_is_set =  sync_event_mock

        # Connection must be None first:
        testrabbit._AsynchronousRabbitConnector__thread._connection = None

        # Run code to be tested:
        testrabbit.send_message_to_queue('foo')

        # Check if the event was unblocked:
        sync_event_mock.wait.assert_called()

        # Check that publish was called:
        feedermock.publish_message.assert_called()

        # Check that the four messages were put into the queue:
        msg_queue = testrabbit._AsynchronousRabbitConnector__unpublished_messages_queue
        self.assert_messages_are_in_queue(msg_queue, ['foo'])
